from pathlib import Path
from pytest import raises
import qrcode
from fpdf.fpdf import FPDF
from fpdf.template import FlexTemplate
from ..conftest import assert_pdf_equal
from test.template.charwrap_test_elements import elements as charwrap_elements

HERE = Path(__file__).resolve().parent


def test_flextemplate_offset(tmp_path):
    elements = [
        {
            "name": "box",
            "type": "B",
            "x1": 0,
            "y1": 0,
            "x2": 50,
            "y2": 50,
        },
        {
            "name": "d1",
            "type": "L",
            "x1": 0,
            "y1": 0,
            "x2": 50,
            "y2": 50,
        },
        {
            "name": "d2",
            "type": "L",
            "x1": 0,
            "y1": 50,
            "x2": 50,
            "y2": 0,
        },
        {
            "name": "label",
            "type": "T",
            "x1": 0,
            "y1": 52,
            "x2": 50,
            "y2": 57,
            "text": "Label",
        },
    ]
    pdf = FPDF()
    pdf.add_page()
    templ = FlexTemplate(pdf, elements)
    templ["label"] = "Offset: 50 / 50 mm"
    templ.render(offsetx=50, offsety=50)
    templ["label"] = "Offset: 50 / 120 mm"
    templ.render(offsetx=50, offsety=120)
    templ["label"] = "Offset: 120 / 50 mm, Scale: 0.5"
    templ.render(offsetx=120, offsety=50, scale=0.5)
    templ["label"] = "Offset: 120 / 120 mm, Rotate: 30°, Scale: 0.5"
    templ.render(offsetx=120, offsety=120, rotate=30.0, scale=0.5)
    assert_pdf_equal(pdf, HERE / "flextemplate_offset.pdf", tmp_path)


def test_flextemplate_multipage(tmp_path):
    elements = [
        {
            "name": "box",
            "type": "B",
            "x1": 0,
            "y1": 0,
            "x2": 50,
            "y2": 50,
        },
        {
            "name": "d1",
            "type": "L",
            "x1": 0,
            "y1": 0,
            "x2": 50,
            "y2": 50,
        },
        {
            "name": "d2",
            "type": "L",
            "x1": 0,
            "y1": 50,
            "x2": 50,
            "y2": 0,
        },
        {
            "name": "e",
            "type": "E",
            "x1": 0,
            "y1": 50,
            "x2": 50,
            "y2": 0,
        },
        {
            "name": "label",
            "type": "T",
            "x1": 0,
            "y1": 52,
            "x2": 50,
            "y2": 57,
            "text": "Label",
        },
    ]
    pdf = FPDF()
    pdf.add_page()
    tmpl_0 = FlexTemplate(pdf, elements)
    tmpl_0["label"] = "Offset: 50 / 50 mm"
    tmpl_0.render(offsetx=50, offsety=50)
    tmpl_0["label"] = "Offset: 50 / 120 mm"
    tmpl_0.render(offsetx=50, offsety=120)
    tmpl_0["label"] = "Offset: 120 / 50 mm"
    tmpl_0.render(offsetx=120, offsety=50)
    tmpl_0["label"] = "Offset: 120 / 120 mm"
    tmpl_0.render(offsetx=120, offsety=120, rotate=30.0)
    pdf.add_page()
    tmpl_0["label"] = "Offset: 120 / 50 mm"
    tmpl_0.render(offsetx=120, offsety=50)
    tmpl_0["label"] = "Offset: 120 / 120 mm"
    tmpl_0.render(offsetx=120, offsety=120, rotate=30.0)
    tmpl_1 = FlexTemplate(pdf)
    tmpl_1.parse_csv(HERE / "template_definition.csv", delimiter=";")
    tmpl_1.render()
    assert_pdf_equal(pdf, HERE / "flextemplate_multipage.pdf", tmp_path)


def test_flextemplate_multipage_parse_json(tmp_path):
    pdf = FPDF()
    tmpl = FlexTemplate(pdf)
    tmpl.parse_json(HERE / "template_definition.json")
    # Repeat on 5 pages, changing a single setting for each page:
    for n in range(1, 6):
        pdf.add_page()
        tmpl["numeric_text"] = f"{n:03d}"
        tmpl.render()
    assert_pdf_equal(pdf, HERE / "flextemplate_multipage_parse_json.pdf", tmp_path)


def test_flextemplate_rotation(tmp_path):
    elements = [
        {
            "name": "box",
            "type": "B",
            "x1": 30,
            "y1": 0,
            "x2": 80,
            "y2": 20,
            "rotate": 10.0,
        },
        {
            "name": "line",
            "type": "L",
            "x1": 0,
            "y1": 0,
            "x2": 50,
            "y2": 20,
            "rotate": 15.0,
        },
        {
            "name": "text!",
            "type": "T",
            "x1": 40,
            "y1": 10,
            "x2": 60,
            "y2": 15,
            "text": "rotatapalooza!",
            "rotate": -15.0,
        },
        {
            "name": "multi",
            "type": "T",
            "x1": 80,
            "y1": 10,
            "x2": 100,
            "y2": 15,
            "text": "Lorem ipsum dolor sit amet, consectetur adipisici elit",
            "rotate": 90.0,
            "background": 0xEEFFFF,
            "multiline": True,
        },
        {
            "name": "write",
            "type": "W",
            "x1": 35,
            "y1": 0,
            "x2": 45,
            "y2": 5,
            "text": "Writing",
            "rotate": -30.0,
        },
        {
            "name": "barcode",
            "type": "BC",
            "x1": 60,
            "y1": 00,
            "y2": 10,
            "text": "123456",
            "size": 1,
            "rotate": 30.0,
        },
        {
            "name": "code39",
            "type": "C39",
            "x1": 80,
            "y1": 10,
            "y2": 15,
            "text": "*987*",
            "size": 1,
            "rotate": 60.0,
        },
        {
            "name": "qrcode",
            "type": "I",
            "x1": 20,
            "y1": 0,
            "x2": 30,
            "y2": 10,
            "rotate": 45,
        },
        {
            "name": "ellipse",
            "type": "E",
            "x1": 45,
            "y1": -10,
            "x2": 65,
            "y2": 0,
            "background": 0x88FFFF,
            "rotate": -45,
        },
    ]
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("courier", size=10)
    templ = FlexTemplate(pdf, elements)
    templ["qrcode"] = qrcode.make("Test 0").get_image()
    templ.render(offsetx=100, offsety=100, rotate=5)
    pdf.add_page()
    scale = 1.2
    for i in range(0, 360, 6):
        templ["qrcode"] = qrcode.make("Test 0").get_image()
        templ.render(offsetx=100, offsety=130, rotate=i, scale=scale)
        scale -= 0.01
    assert_pdf_equal(pdf, HERE / "flextemplate_rotation.pdf", tmp_path)


def test_flextemplate_badinput():
    with raises(TypeError):
        FlexTemplate("NotAnFPDF()Instance")


def test_flextemplate_elements(tmp_path):
    """Check that all elements end up in the right place."""
    grid_elements = (
        dict(name="v02", type="L", x1=20, y1=20, x2=20, y2=280),
        dict(name="v04", type="L", x1=40, y1=20, x2=40, y2=280),
        dict(name="v06", type="L", x1=60, y1=20, x2=60, y2=280),
        dict(name="v08", type="L", x1=80, y1=20, x2=80, y2=280),
        dict(name="v10", type="L", x1=100, y1=20, x2=100, y2=280),
        dict(name="v12", type="L", x1=120, y1=20, x2=120, y2=280),
        dict(name="v14", type="L", x1=140, y1=20, x2=140, y2=280),
        dict(name="v15", type="L", x1=160, y1=20, x2=160, y2=280),
        dict(name="v18", type="L", x1=180, y1=20, x2=180, y2=280),
        dict(name="h02", type="L", x1=20, y1=20, x2=180, y2=20),
        dict(name="h04", type="L", x1=20, y1=40, x2=180, y2=40),
        dict(name="h06", type="L", x1=20, y1=60, x2=180, y2=60),
        dict(name="h08", type="L", x1=20, y1=80, x2=180, y2=80),
        dict(name="h10", type="L", x1=20, y1=100, x2=180, y2=100),
        dict(name="h12", type="L", x1=20, y1=120, x2=180, y2=120),
        dict(name="h14", type="L", x1=20, y1=140, x2=180, y2=140),
        dict(name="h16", type="L", x1=20, y1=160, x2=180, y2=160),
        dict(name="h18", type="L", x1=20, y1=180, x2=180, y2=180),
        dict(name="h20", type="L", x1=20, y1=200, x2=180, y2=200),
        dict(name="h22", type="L", x1=20, y1=220, x2=180, y2=220),
        dict(name="h24", type="L", x1=20, y1=240, x2=180, y2=240),
        dict(name="h26", type="L", x1=20, y1=260, x2=180, y2=260),
        dict(name="h28", type="L", x1=20, y1=280, x2=180, y2=280),
    )
    text_elements = (
        dict(
            name="t",
            type="T",
            x1=0,
            y1=0,
            x2=10,
            y2=4,
            text="text0",
            background=0xFFFF00,
        ),
        dict(
            name="t2",
            type="T",
            x1=20,
            y1=0,
            x2=30,
            y2=4,
            text="text2",
            background=0xFFBB00,
        ),
        dict(
            name="t3",
            type="T",
            x1=00,
            y1=30,
            x2=10,
            y2=34,
            text="text3",
            background=0xBBBF00,
            rotate=10,
        ),
    )
    ml_elements = (
        dict(
            name="ml",
            type="T",
            x1=0,
            y1=0,
            x2=15,
            y2=4,
            text="Lorem ipsum dolor sit amet",
            multiline=True,
        ),
        dict(
            name="ml2",
            type="T",
            x1=20,
            y1=0,
            x2=35,
            y2=4,
            text="Lorem ipsum dolor sit amet",
            multiline=True,
        ),
        dict(
            name="ml3",
            type="T",
            x1=00,
            y1=30,
            x2=15,
            y2=34,
            text="Lorem ipsum dolor sit amet",
            multiline=True,
            rotate=10,
        ),
    )
    write_elements = (
        dict(
            name="w",
            type="W",
            x1=0,
            y1=0,
            x2=10,
            y2=4,
            text="write0",
            background=0xFFFF00,
        ),
        dict(
            name="w2",
            type="W",
            x1=20,
            y1=0,
            x2=30,
            y2=4,
            text="write2",
            background=0xFFBB00,
        ),
        dict(
            name="w3",
            type="W",
            x1=00,
            y1=30,
            x2=10,
            y2=34,
            text="write3",
            background=0xBBBF00,
            rotate=10,
        ),
    )
    line_elements = (
        dict(name="l", type="L", x1=0, y1=0, x2=10, y2=0, size=2),
        dict(name="l2", type="L", x1=20, y1=0, x2=30, y2=0, size=2),
        dict(name="l3", type="L", x1=00, y1=30, x2=10, y2=30, size=2, rotate=10),
    )
    box_elements = (
        dict(name="b", type="B", x1=0, y1=0, x2=10, y2=4),
        dict(name="b2", type="B", x1=20, y1=0, x2=30, y2=4),
        dict(name="b3", type="B", x1=00, y1=30, x2=10, y2=34, rotate=10),
    )
    ellipse_elements = (
        dict(name="e", type="E", x1=0, y1=0, x2=10, y2=4),
        dict(name="e2", type="E", x1=20, y1=0, x2=30, y2=4),
        dict(name="e3", type="E", x1=00, y1=30, x2=10, y2=34, rotate=10),
    )
    bc_elements = (
        dict(name="bc", type="BC", x1=0, y1=0, x2=10, y2=4, text="98", size=1),
        dict(name="bc2", type="BC", x1=20, y1=0, x2=30, y2=4, text="01", size=1),
        dict(
            name="bc3",
            type="BC",
            x1=00,
            y1=30,
            x2=10,
            y2=34,
            text="01",
            size=1,
            rotate=10,
        ),
    )
    c39_elements = (
        dict(name="c39", type="C39", x1=0, y1=0, x2=10, y2=4, text="*xy*", size=0.7),
        dict(name="c39_2", type="C39", x1=20, y1=0, x2=30, y2=4, text="*01*", size=0.7),
        dict(
            name="c39_3",
            type="C39",
            x1=00,
            y1=30,
            x2=10,
            y2=34,
            text="*01*",
            size=0.7,
            rotate=10,
        ),
    )
    img_elements = (
        dict(name="i", type="I", x1=0, y1=0, x2=10, y2=10, size=0.7),
        dict(name="i2", type="I", x1=20, y1=0, x2=30, y2=10, size=0.7),
        dict(name="i3", type="I", x1=00, y1=30, x2=10, y2=40, size=0.7, rotate=10),
    )
    pdf = FPDF()
    grid_t = FlexTemplate(pdf, grid_elements)
    pdf.add_page()
    grid_t.render()
    text_t = FlexTemplate(pdf, text_elements)
    text_t.render(offsetx=40, offsety=40)
    text_t.render(offsetx=40, offsety=100, rotate=30)
    text_t.render(offsetx=80, offsety=200, rotate=45)
    pdf.add_page()
    grid_t.render()
    ml_t = FlexTemplate(pdf, ml_elements)
    ml_t.render(offsetx=40, offsety=40)
    ml_t.render(offsetx=40, offsety=100, rotate=30)
    ml_t.render(offsetx=80, offsety=200, rotate=45)
    pdf.add_page()
    grid_t.render()
    write_t = FlexTemplate(pdf, write_elements)
    write_t.render(offsetx=40, offsety=40)
    write_t.render(offsetx=40, offsety=100, rotate=30)
    write_t.render(offsetx=80, offsety=200, rotate=45)
    pdf.add_page()
    grid_t.render()
    line_t = FlexTemplate(pdf, line_elements)
    line_t.render(offsetx=40, offsety=40)
    line_t.render(offsetx=40, offsety=100, rotate=30)
    line_t.render(offsetx=80, offsety=200, rotate=45)
    pdf.add_page()
    grid_t.render()
    box_t = FlexTemplate(pdf, box_elements)
    box_t.render(offsetx=40, offsety=40)
    box_t.render(offsetx=40, offsety=100, rotate=30)
    box_t.render(offsetx=80, offsety=200, rotate=45)
    pdf.add_page()
    grid_t.render()
    ell_t = FlexTemplate(pdf, ellipse_elements)
    ell_t.render(offsetx=40, offsety=40)
    ell_t.render(offsetx=40, offsety=100, rotate=30)
    ell_t.render(offsetx=80, offsety=200, rotate=45)

    pdf.add_page()
    grid_t.render()
    bc_t = FlexTemplate(pdf, bc_elements)
    bc_t.render(offsetx=40, offsety=40)
    bc_t.render(offsetx=40, offsety=100, rotate=30)
    bc_t.render(offsetx=80, offsety=200, rotate=45)

    pdf.add_page()
    grid_t.render()
    c39_t = FlexTemplate(pdf, c39_elements)
    c39_t.render(offsetx=40, offsety=40)
    c39_t.render(offsetx=40, offsety=100, rotate=30)
    c39_t.render(offsetx=80, offsety=200, rotate=45)

    pdf.add_page()
    grid_t.render()
    img = qrcode.make("Test 0").get_image()
    img_t = FlexTemplate(pdf, img_elements)
    img_t["i"] = img
    img_t["i2"] = img
    img_t["i3"] = img
    img_t.render(offsetx=40, offsety=40)
    img_t["i"] = img
    img_t["i2"] = img
    img_t["i3"] = img
    img_t.render(offsetx=40, offsety=100, rotate=30)
    img_t["i"] = img
    img_t["i2"] = img
    img_t["i3"] = img
    img_t.render(offsetx=80, offsety=200, rotate=45)
    assert_pdf_equal(pdf, HERE / "flextemplate_elements.pdf", tmp_path)


def test_flextemplate_leak(tmp_path):  # issue #570
    # rendering a template should not change the current graphics context.
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("helvetica", size=10)
    pdf.cell(text="before", new_x="LEFT", new_y="NEXT")
    tmpldata = [
        {
            "name": "template",
            "type": "T",
            "x1": 0,
            "y1": 0,
            "x2": 50,
            "y2": 10,
            "bold": True,
            "underline": True,
            "text": "Template",
        },
    ]
    templ = FlexTemplate(pdf, elements=tmpldata)
    templ.render(offsetx=pdf.x, offsety=pdf.y)
    pdf.ln()
    pdf.cell(text="after", new_x="LEFT", new_y="NEXT")
    assert_pdf_equal(pdf, HERE / "flextemplate_leak.pdf", tmp_path)


def test_flextemplate_wrapmode(tmp_path):
    """Test that wrap mode can optionally be used to set wrapping using characters instead of words"""

    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("courier", size=10)
    templ = FlexTemplate(pdf, charwrap_elements)
    templ.render()
    assert_pdf_equal(pdf, HERE / "flextemplate_wrapmode.pdf", tmp_path)


def test_flextemplate_dashed_elements(tmp_path):
    """Test that elements can be created with dashed lines and that the dash pattern does not leak."""

    elements = [
        {
            "name": "solid-line1",
            "type": "L",
            "x1": 10,
            "y1": 5,
            "x2": 100,
            "y2": 5,
            "size": 1,
        },
        {
            "name": "dashed-line",
            "type": "L",
            "x1": 10,
            "y1": 10,
            "x2": 100,
            "y2": 10,
            "size": 1,
            "dash_pattern": {"dash": 2, "gap": 4, "phase": 0.5},
        },
        {
            "name": "solid-line2",
            "type": "L",
            "x1": 10,
            "y1": 15,
            "x2": 100,
            "y2": 15,
            "size": 1,
        },
        {
            "name": "solid-rect1",
            "type": "B",
            "x1": 5,
            "y1": 25,
            "x2": 105,
            "y2": 55,
            "size": 1.5,
        },
        {
            "name": "dashed-rect",
            "type": "B",
            "x1": 10,
            "y1": 30,
            "x2": 100,
            "y2": 50,
            "size": 0.5,
            "dash_pattern": {"dash": 2},
        },
        {
            "name": "solid-rect2",
            "type": "B",
            "x1": 15,
            "y1": 35,
            "x2": 95,
            "y2": 45,
            "size": 0.5,
        },
        {
            "name": "solid-ellipse1",
            "type": "E",
            "x1": 15,
            "y1": 85,
            "x2": 95,
            "y2": 95,
            "size": 0.5,
        },
        {
            "name": "dashed-ellipse",
            "type": "E",
            "x1": 10,
            "y1": 80,
            "x2": 100,
            "y2": 100,
            "size": 0.5,
            "dash_pattern": {"dash": 2, "gap": 1},
        },
        {
            "name": "solid-ellipse2",
            "type": "E",
            "x1": 5,
            "y1": 75,
            "x2": 105,
            "y2": 105,
            "size": 0.5,
        },
    ]

    pdf = FPDF()
    template = FlexTemplate(pdf, elements)

    # First page has a solid dash_pattern before rendering the template.
    pdf.add_page()
    pdf.set_dash_pattern()
    template.render()

    # Second page has a dash_pattern applied prior to rendering the template.
    pdf.add_page()
    pdf.set_dash_pattern(5, 5, 1)
    template.render()

    assert_pdf_equal(pdf, HERE / "flextemplate_dashed_elements.pdf", tmp_path)


def test_flextemplate_dash_pattern_badinput():
    elements = [
        {
            "name": "bad-line",
            "type": "L",
            "x1": 100,
            "y1": 100,
            "x2": 200,
            "y2": 150,
            "dash_pattern": {"dash": 5, "bad_key": 6},
        }
    ]

    pdf = FPDF()
    with raises(KeyError):
        template = FlexTemplate(pdf, elements)
        template.render()



import pytest
from pathlib import Path

from fpdf import FPDF
from fpdf.drawing import PaintedPath, PaintComposite
from fpdf.enums import CompositingOperation, PathPaintRule
from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent


def _blue_square():
    p = PaintedPath()
    p.rectangle(10, 10, 50, 50)
    p.style.fill_color = "#0000ff"
    p.style.fill_opacity = 1
    p.style.paint_rule = PathPaintRule.FILL_NONZERO
    return p


def _red_square():
    p = PaintedPath()
    p.rectangle(35, 35, 50, 50)
    p.style.fill_color = "#ff0000"
    p.style.fill_opacity = 1
    p.style.paint_rule = PathPaintRule.FILL_NONZERO
    return p


def _generate_pdf(op: CompositingOperation):
    pdf = FPDF()
    pdf.page_background = (211, 211, 211)
    pdf.add_page()
    with pdf.drawing_context() as gc:
        src = _blue_square()
        dst = _red_square()
        comp = PaintComposite(backdrop=dst, source=src, operation=op)
        gc.add_item(comp)
    return pdf


@pytest.mark.parametrize(
    "op",
    [
        CompositingOperation.CLEAR,
        CompositingOperation.SOURCE,
        CompositingOperation.DESTINATION,
        CompositingOperation.SOURCE_OVER,
        CompositingOperation.DESTINATION_OVER,
        CompositingOperation.SOURCE_IN,
        CompositingOperation.DESTINATION_IN,
        CompositingOperation.SOURCE_OUT,
        CompositingOperation.DESTINATION_OUT,
        CompositingOperation.SOURCE_ATOP,
        CompositingOperation.DESTINATION_ATOP,
        CompositingOperation.XOR,
    ],
)
def test_compositing_operations(op, tmp_path):
    pdf = _generate_pdf(op)
    name = f"compositing_{op.value}.pdf"
    assert_pdf_equal(
        pdf,
        HERE / "generated_pdf" / name,
        tmp_path,
    )



#!/usr/bin/env python3
# USAGE: ./add_new_page_with_pdfrw.py src_file.pdf dest_file.pdf
import sys

from fpdf import FPDF
from pdfrw import PdfReader, PdfWriter

IN_FILEPATH = sys.argv[1]
OUT_FILEPATH = sys.argv[2]
NEW_PAGE_INDEX = 1  # set to None to append at the end


def new_page():
    fpdf = FPDF()
    fpdf.add_page()
    fpdf.set_font("helvetica", size=36)
    fpdf.text(50, 50, "Hello!")
    reader = PdfReader(fdata=bytes(fpdf.output()))
    return reader.pages[0]


writer = PdfWriter(trailer=PdfReader(IN_FILEPATH))
writer.addpage(new_page(), at_index=NEW_PAGE_INDEX)
writer.write(OUT_FILEPATH)



from pathlib import Path

import qrcode, pytest

from fpdf import FPDF
from test.conftest import assert_pdf_equal, LOREM_IPSUM

HERE = Path(__file__).resolve().parent
IMG_DIR = HERE.parent / "image"

TABLE_DATA = (
    ("First name", "Last name", "Image", "City"),
    (
        "Jules",
        "Smith",
        IMG_DIR / "png_images/ba2b2b6e72ca0e4683bb640e2d5572f8.png",
        "San Juan",
    ),
    (
        "Mary",
        "Ramos",
        IMG_DIR / "png_images/ac6343a98f8edabfcc6e536dd75aacb0.png",
        "Orlando",
    ),
    (
        "Carlson",
        "Banks",
        IMG_DIR / "image_types/insert_images_insert_png.png",
        "Los Angeles",
    ),
    ("Lucas", "Cimon", IMG_DIR / "image_types/circle.bmp", "Angers"),
)
MULTILINE_TABLE_DATA = (
    ("Multilines text", "Image"),
    (LOREM_IPSUM[:200], IMG_DIR / "png_images/ba2b2b6e72ca0e4683bb640e2d5572f8.png"),
    (LOREM_IPSUM[200:400], IMG_DIR / "png_images/ac6343a98f8edabfcc6e536dd75aacb0.png"),
    (LOREM_IPSUM[400:600], IMG_DIR / "image_types/insert_images_insert_png.png"),
    (LOREM_IPSUM[600:800], IMG_DIR / "image_types/circle.bmp"),
)


def test_table_with_image_border_overlap(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    with pdf.table() as table:
        row = table.row()
        row.cell(img=IMG_DIR / "png_indexed/flower1.png", img_fill_width=True)
        row.cell("Other field")
    assert_pdf_equal(pdf, HERE / "table_with_image_border_overlap.pdf", tmp_path)


def test_table_with_image_mixed_rows_and_alignment(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    with pdf.table() as table:
        for i, data_row in enumerate(MULTILINE_TABLE_DATA[:2]):
            row = table.row()
            for j, datum in enumerate(data_row):
                if j == 1 and i > 0:
                    row.cell(img=datum, img_fill_width=True)
                else:
                    row.cell(datum)

        row = table.row()
        row.cell("Where do I align vertically?")
        row.cell("This is a cell with multiple lines\n" * 6)

    assert_pdf_equal(
        pdf,
        HERE / "table_with_multiline_cells_and_images_mixed.pdf",
        tmp_path,
    )


def test_table_with_images(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    with pdf.table() as table:
        for i, data_row in enumerate(TABLE_DATA):
            row = table.row()
            for j, datum in enumerate(data_row):
                if j == 2 and i > 0:
                    row.cell(img=datum)
                else:
                    row.cell(datum)
    assert_pdf_equal(pdf, HERE / "table_with_images.pdf", tmp_path)


def test_table_with_images_and_img_fill_width(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    with pdf.table() as table:
        for i, data_row in enumerate(TABLE_DATA):
            row = table.row()
            for j, datum in enumerate(data_row):
                if j == 2 and i > 0:
                    row.cell(img=datum, img_fill_width=True)
                else:
                    row.cell(datum)
    assert_pdf_equal(pdf, HERE / "table_with_images_and_img_fill_width.pdf", tmp_path)


def test_table_with_multiline_cells_and_images(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    with pdf.table() as table:
        for i, data_row in enumerate(MULTILINE_TABLE_DATA):
            row = table.row()
            for j, datum in enumerate(data_row):
                if j == 1 and i > 0:
                    row.cell(img=datum, img_fill_width=True)
                else:
                    row.cell(datum)
    assert_pdf_equal(pdf, HERE / "table_with_multiline_cells_and_images.pdf", tmp_path)


def test_table_with_images_and_text():
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    with pytest.raises(NotImplementedError):
        with pdf.table() as table:
            for i, data_row in enumerate(TABLE_DATA):
                row = table.row()
                for j, datum in enumerate(data_row):
                    if j == 2 and i > 0:
                        row.cell(datum.name, img=datum)
                    else:
                        row.cell(datum)


def test_table_with_qrcode(tmp_path):  # issue 771
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    qrCodeGenerated = qrcode.make("https://google.com")
    with pdf.table(first_row_as_headings=False) as table:
        row = table.row()
        row.cell(img=qrCodeGenerated.get_image(), img_fill_width=True)
        row.cell("Other field")
    assert_pdf_equal(pdf, HERE / "table_with_qrcode.pdf", tmp_path)


def test_table_with_page_break_over_image_with_heading(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    pdf.y = 220
    with pdf.table() as table:
        for i, data_row in enumerate(TABLE_DATA[:2]):
            row = table.row()
            for j, datum in enumerate(data_row):
                if j == 2 and i > 0:
                    row.cell(img=datum, img_fill_width=True)
                else:
                    row.cell(datum)
    assert_pdf_equal(
        pdf, HERE / "table_with_page_break_over_image_with_heading.pdf", tmp_path
    )


def test_table_with_page_break_over_image_without_heading(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    pdf.y = 220
    with pdf.table(first_row_as_headings=False) as table:
        for i, data_row in enumerate(TABLE_DATA[:2]):
            row = table.row()
            for j, datum in enumerate(data_row):
                if j == 2 and i > 0:
                    row.cell(img=datum, img_fill_width=True)
                else:
                    row.cell(datum)
    assert_pdf_equal(
        pdf, HERE / "table_with_page_break_over_image_without_heading.pdf", tmp_path
    )


def test_table_with_images_and_links(tmp_path):
    pdf = FPDF()
    pdf.set_font("Times", size=16)
    pdf.add_page()
    pdf.y = 100
    pdf.cell(text="Page 1", center=True)
    pdf.add_page()
    with pdf.table() as table:
        for i, data_row in enumerate(TABLE_DATA):
            row = table.row()
            for j, datum in enumerate(data_row):
                if j == 2 and i > 0:
                    row.cell(
                        img=datum,
                        img_fill_width=True,
                        link="https://py-pdf.github.io/fpdf2/",
                    )
                else:
                    row.cell(datum, link=pdf.add_link(page=1))
    assert_pdf_equal(pdf, HERE / "table_with_images_and_links.pdf", tmp_path)



#!/usr/bin/env python3
# USAGE: ./add_new_page_with_pypdf.py src_file.pdf dest_file.pdf
import io, sys

from fpdf import FPDF
from pypdf import PdfReader, PdfWriter

IN_FILEPATH = sys.argv[1]
OUT_FILEPATH = sys.argv[2]
ON_PAGE_INDEX = 2  # Index at which the page will be inserted (starts at zero)


def build_page():
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("times", style="B", size=19)
    pdf.text(50, 10, "Hello World!")
    return io.BytesIO(pdf.output())


writer = PdfWriter(clone_from=IN_FILEPATH)
new_page = PdfReader(build_page()).pages[0]
writer.insert_page(new_page, index=ON_PAGE_INDEX)
writer.write(OUT_FILEPATH)



import gzip
from pathlib import Path

import pytest

from fpdf import FPDF, FPDFException
from test.conftest import (
    assert_pdf_equal,
    assert_same_file,
    ensure_exec_time_below,
    ensure_rss_memory_below,
    LOREM_IPSUM,
)

HERE = Path(__file__).resolve().parent
FONTS_DIR = HERE.parent / "fonts"

with gzip.open(HERE / "long_text.txt.gz") as text_file:
    LONG_TEXT_LINES = text_file.read().decode(encoding="utf-8").splitlines()

TEXT_SIZE, SPACING = 36, 1.15
LINE_HEIGHT = TEXT_SIZE * SPACING

TABLE_DATA = (
    ("First name", "Last name", "Age", "City"),
    ("Jules", "Smith", "34", "San Juan"),
    ("Mary", "Ramos", "45", "Orlando"),
    ("Carlson", "Banks", "19", "Los Angeles"),
    ("Lucas", "Cimon", "31", "Angers"),
)


def test_ln_positioning_and_page_breaking_for_cell(tmp_path):
    doc = FPDF(format="letter", unit="pt")
    doc.add_page()
    doc.set_font("helvetica", size=TEXT_SIZE)
    text = LOREM_IPSUM * 100
    for i in range(20):
        doc.cell(
            w=72,
            h=TEXT_SIZE * 1.5,
            border=1,
            new_x="LEFT",
            new_y="NEXT",
            text=text[i * 100 : i * 100 + 99],
        )

    assert_pdf_equal(
        doc, HERE / "ln_positioning_and_page_breaking_for_cell.pdf", tmp_path
    )


def test_cell_ln_0(tmp_path):
    doc = FPDF()
    doc.add_page()
    doc.set_font("helvetica", size=TEXT_SIZE)
    doc.cell(w=45, h=LINE_HEIGHT, border=1, text="Lorem")
    doc.cell(w=45, h=LINE_HEIGHT, border=1, text="ipsum")
    doc.cell(w=45, h=LINE_HEIGHT, border=1, text="Ut")
    doc.cell(w=45, h=LINE_HEIGHT, border=1, text="nostrud")
    assert_pdf_equal(doc, HERE / "ln_0.pdf", tmp_path)


def test_cell_ln_1(tmp_path):
    """
    Catch DeprecationWarning for ln=1.
    Validating that: "Using ln=1 is equivalent to using ln=0 and calling the `ln` method just after."
    """
    doc = FPDF()
    doc.add_page()
    doc.set_font("helvetica", size=TEXT_SIZE)
    with pytest.warns(DeprecationWarning) as record:
        doc.cell(w=100, h=LINE_HEIGHT, border=1, text="Lorem ipsum", ln=1)
    assert len(record) == 1
    assert_same_file(record[0].filename, __file__)

    doc.cell(w=100, h=LINE_HEIGHT, border=1, text="Ut nostrud irure")
    assert_pdf_equal(doc, HERE / "ln_1.pdf", tmp_path)

    doc = FPDF()
    doc.add_page()
    doc.set_font("helvetica", size=TEXT_SIZE)
    doc.cell(w=100, h=LINE_HEIGHT, border=1, text="Lorem ipsum")
    doc.ln()
    doc.cell(w=100, h=LINE_HEIGHT, border=1, text="Ut nostrud irure")
    assert_pdf_equal(doc, HERE / "ln_1.pdf", tmp_path)


def test_cell_table_with_pagebreak(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    line_height = pdf.font_size * 2
    col_width = pdf.epw / 4  # distribute content evenly
    for i in range(4):  # repeat table 4 times
        for row in TABLE_DATA:
            for datum in row:
                pdf.cell(col_width, line_height, f"{datum} ({i})", border=1)
            pdf.ln(line_height)
        pdf.ln(line_height * 2)
    assert_pdf_equal(pdf, HERE / "cell_table_with_pagebreak.pdf", tmp_path)


def test_cell_table_unbreakable(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    line_height = pdf.font_size * 2
    col_width = pdf.epw / 4  # distribute content evenly
    for i in range(5):  # repeat table 5 times
        with pdf.unbreakable() as doc:
            for row in TABLE_DATA:
                for datum in row:
                    doc.cell(col_width, line_height, f"{datum} ({i})", border=1)
                doc.ln(line_height)
        pdf.ln(line_height * 2)
    assert_pdf_equal(pdf, HERE / "cell_table_unbreakable.pdf", tmp_path)


def test_cell_without_font_set():
    pdf = FPDF()
    pdf.add_page()
    with pytest.raises(FPDFException) as error:
        pdf.cell(text="Hello World!")
    expected_msg = "No font set, you need to call set_font() beforehand"
    assert str(error.value) == expected_msg


def test_cell_without_w_nor_h(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    pdf.cell(text="Lorem ipsum", border=1)
    pdf.set_font_size(80)
    pdf.cell(text="Lorem ipsum", border=1)
    assert_pdf_equal(pdf, HERE / "cell_without_w_nor_h.pdf", tmp_path)


def test_cell_missing_text_or_width():
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    with pytest.raises(ValueError) as error:
        pdf.cell()
    assert (
        str(error.value)
        == "'text_line' must have fragments if 'text_line.text_width' is None"
    )


def test_cell_centering(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=60)
    pdf.cell(text="Lorem ipsum", border=1, center=True)
    assert_pdf_equal(pdf, HERE / "cell_centering.pdf", tmp_path)


def test_cell_centering_and_align_x(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    pdf.cell(text="Lorem ipsum", border=1, center=True, align="X")
    pdf.set_draw_color(r=0, g=255, b=0)
    pdf.line(pdf.w / 2, 0, pdf.w / 2, pdf.h)
    assert_pdf_equal(pdf, HERE / "cell_centering_and_align_x.pdf", tmp_path)


def test_cell_markdown(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=60)
    pdf.cell(text="**Lorem** __Ipsum__ --dolor--", markdown=True)
    assert_pdf_equal(pdf, HERE / "cell_markdown.pdf", tmp_path)


def test_cell_markdown_escaped(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=40)
    pdf.cell(text="**Lo\\rem** \\__Ipsum\\__ \\\\--dolor\\\\--", markdown=True)
    pdf.write(text="\n")
    pdf.cell(text="\\****BOLD**\\**", markdown=True)
    assert_pdf_equal(pdf, HERE / "cell_markdown_escaped.pdf", tmp_path)


def test_cell_markdown_bold_italic(tmp_path):
    # issue 1094
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=60)
    pdf.cell(text="**__Lorem --Ipsum--__**", markdown=True)
    assert_pdf_equal(pdf, HERE / "cell_markdown_bold_italic.pdf", tmp_path)


def test_cell_markdown_bold_italic_escaped(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=40)
    pdf.cell(text="**__Lorem \\--Ipsum\\--__**", markdown=True)
    assert_pdf_equal(pdf, HERE / "cell_markdown_bold_italic_escaped.pdf", tmp_path)


def test_cell_markdown_with_ttf_fonts(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.add_font("Roboto", "", FONTS_DIR / "Roboto-Regular.ttf")
    pdf.add_font("Roboto", "B", FONTS_DIR / "Roboto-Bold.ttf")
    pdf.add_font("Roboto", "I", FONTS_DIR / "Roboto-Italic.ttf")
    pdf.set_font("Roboto", size=60)
    pdf.cell(text="**Lorem** __Ipsum__ --dolor--", markdown=True)
    assert_pdf_equal(pdf, HERE / "cell_markdown_with_ttf_fonts.pdf", tmp_path)


def test_cell_markdown_with_ttf_fonts_escaped(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.add_font("Roboto", "", FONTS_DIR / "Roboto-Regular.ttf")
    pdf.add_font("Roboto", "B", FONTS_DIR / "Roboto-Bold.ttf")
    pdf.add_font("Roboto", "I", FONTS_DIR / "Roboto-Italic.ttf")
    pdf.set_font("Roboto", size=40)
    pdf.cell(text="**Lo\\rem** \\__Ipsum\\__ \\\\--dolor\\\\--", markdown=True)
    assert_pdf_equal(pdf, HERE / "cell_markdown_with_ttf_fonts_escaped.pdf", tmp_path)


def test_cell_markdown_missing_ttf_font():
    pdf = FPDF()
    pdf.add_page()
    pdf.add_font(fname=FONTS_DIR / "Roboto-Regular.ttf")
    pdf.set_font("Roboto-Regular", size=60)
    with pytest.raises(FPDFException) as error:
        pdf.cell(text="**Lorem Ipsum**", markdown=True)
    expected_msg = "Undefined font: roboto-regularB - Use built-in fonts or FPDF.add_font() beforehand"
    assert str(error.value) == expected_msg


def test_cell_markdown_bleeding(tmp_path):  # issue 241
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=60)
    pdf.cell(text="**Lorem Ipsum dolor**", markdown=True, new_x="LMARGIN", new_y="NEXT")
    assert pdf.font_style == ""
    pdf.cell(text="No Markdown", markdown=False, new_x="LMARGIN", new_y="NEXT")
    pdf.cell(text="--Lorem Ipsum dolor--", markdown=True, new_x="LMARGIN", new_y="NEXT")
    assert pdf.font_style == ""
    pdf.cell(text="No Markdown", markdown=False, new_x="LMARGIN", new_y="NEXT")
    pdf.cell(text="__Lorem Ipsum dolor__", markdown=True, new_x="LMARGIN", new_y="NEXT")
    assert pdf.font_style == ""
    pdf.cell(text="No Markdown", markdown=False, new_x="LMARGIN", new_y="NEXT")
    assert_pdf_equal(pdf, HERE / "cell_markdown_bleeding.pdf", tmp_path)


def test_cell_markdown_right_aligned(tmp_path):  # issue 333
    pdf = FPDF()
    pdf.add_page()
    pdf.add_font("Roboto", fname=FONTS_DIR / "Roboto-Regular.ttf")
    pdf.add_font("Roboto", style="B", fname=FONTS_DIR / "Roboto-Bold.ttf")
    pdf.set_font("Roboto", size=60)
    pdf.cell(
        0,
        9,
        "**X** **X** **X** **X** **X** **X** **X** **X** **X**",
        markdown=True,
        align="R",
    )
    assert_pdf_equal(pdf, HERE / "cell_markdown_right_aligned.pdf", tmp_path)


def test_table_with_headers_on_every_page(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    line_height = pdf.font_size * 2
    col_width = pdf.epw / 4  # distribute content evenly

    def render_table_header():
        pdf.set_font(style="B")  # enabling bold text
        for col_name in TABLE_DATA[0]:
            pdf.cell(col_width, line_height, col_name, border=1)
        pdf.ln(line_height)
        pdf.set_font(style="")  # disabling bold text

    render_table_header()
    for _ in range(10):  # repeat data rows
        for row in TABLE_DATA[1:]:
            if pdf.will_page_break(line_height):
                render_table_header()
            for datum in row:
                pdf.cell(col_width, line_height, datum, border=1)
            pdf.ln(line_height)
    assert_pdf_equal(pdf, HERE / "table_with_headers_on_every_page.pdf", tmp_path)


def test_cell_newpos_badinput():
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    with pytest.raises(ValueError):
        pdf.cell(w=0, ln=5)

    with pytest.raises(TypeError):
        pdf.cell(w=0, new_x=5)
    with pytest.raises(TypeError):
        pdf.cell(w=0, new_y=None)
    with pytest.raises(ValueError):
        pdf.cell(w=0, align="J")


def test_cell_curfont_leak(tmp_path):  # issue #475
    pdf = FPDF()
    pdf.add_page()
    pdf.add_font("Roboto", fname=FONTS_DIR / "Roboto-Regular.ttf")
    pdf.add_font("Roboto", style="B", fname=FONTS_DIR / "Roboto-Bold.ttf")
    with pdf.local_context():
        pdf.set_font("Roboto", style="B", size=10)
        pdf.cell(text="ABCDEFGH", new_x="LEFT", new_y="NEXT")
    pdf.set_font("Roboto", size=10)
    pdf.cell(text="IJKLMNOP", new_x="LEFT", new_y="NEXT")
    with pdf.local_context():
        pdf.set_font("Roboto", style="B", size=10)
        pdf.cell(text="QRSTUVW", new_x="LEFT", new_y="NEXT")
    pdf.set_font("Roboto", size=10)
    pdf.cell(text="XYZ012abc,-", new_x="LEFT", new_y="NEXT")
    pdf.cell(text="3,7E-05", new_x="LEFT", new_y="NEXT")
    assert_pdf_equal(pdf, HERE / "cell_curfont_leak.pdf", tmp_path)


# pylint: disable=protected-access
def test_cell_lasth(tmp_path):  # issue #601
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("helvetica", size=18)
    pdf.set_fill_color(255, 255, 0)
    pdf.cell(w=100, text="Hello world", fill=True)
    pdf.ln()
    assert pdf._lasth == 6.35, f"pdf._lasth ({pdf._lasth}) != 5.35"
    pdf.set_fill_color(255, 0, 255)
    pdf.cell(w=100, text="Hello world", h=50, fill=True)
    pdf.ln()
    assert pdf._lasth == 50, f"pdf._lasth ({pdf._lasth}) != 50 after cell(h=50)"
    pdf.set_fill_color(0, 255, 255)
    pdf.cell(w=100, text="Hello world", fill=True)
    pdf.cell(w=100, text="")
    pdf.ln()
    assert pdf._lasth == 6.35, f"pdf._lasth ({pdf._lasth}) != 5.35 after empty cell"
    pdf.cell(w=100, text="Hello world", border=True)
    assert_pdf_equal(pdf, HERE / "cell_lasth.pdf", tmp_path)


def test_cell_deprecated_txt_arg():
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Helvetica", size=TEXT_SIZE)
    with pytest.warns(
        DeprecationWarning, match='The parameter "txt" has been renamed to "text"'
    ):
        # pylint: disable=unexpected-keyword-arg
        pdf.cell(txt="Lorem ipsum Ut nostrud irure")


@ensure_exec_time_below(seconds=18)
@ensure_rss_memory_below(mib=1)
def test_cell_speed_with_long_text():  # issue #907
    pdf = FPDF()
    pdf.add_font(fname=FONTS_DIR / "DejaVuSans.ttf")
    pdf.set_font("DejaVuSans", size=5)
    pdf.add_page()
    assert len(LONG_TEXT_LINES) == 26862
    for line in LONG_TEXT_LINES:
        pdf.cell(0, 3, line, new_x="LMARGIN", new_y="NEXT")



from pathlib import Path

from fpdf import FPDF
from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent


def test_rotation(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    x, y = 60, 60
    img_filepath = HERE / "image/png_images/66ac49ef3f48ac9482049e1ab57a53e9.png"
    with pdf.rotation(45, x=x, y=y):
        pdf.image(img_filepath, x=x, y=y)
    pdf.image(img_filepath, x=150, y=150)
    assert_pdf_equal(pdf, HERE / "rotation.pdf", tmp_path)



from fpdf import FPDF
from aztec_code_generator import AztecCode

pdf = FPDF()
pdf.add_page()
aztec_code = AztecCode("https://py-pdf.github.io/fpdf2/")
pdf.image(aztec_code.image(), x=10, y=10, w=100, h=100)
pdf.output("aztec_code.pdf")



"""
Quoting section 8.2.2 "Document Outline" of the 2006 PDF spec 1.7:
> The document outline consists of a tree-structured hierarchy of outline items (sometimes called bookmarks),
> which serve as a visual table of contents to display the document’s structure to the user.

The contents of this module are internal to fpdf2, and not part of the public API.
They may change at any time without prior warning or any deprecation period,
in non-backward-compatible ways.

Usage documentation at: <https://py-pdf.github.io/fpdf2/DocumentOutlineAndTableOfContents.html>
"""

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Iterator, Optional, Sequence

from .enums import Align, XPos, YPos
from .fonts import TextStyle
from .structure_tree import StructElem
from .syntax import DestinationXYZ, PDFObject, PDFString
from .util import FloatTolerance

if TYPE_CHECKING:
    from .fpdf import FPDF


@dataclass
class OutlineSection:
    # RAM usage optimization
    __slots__ = ("name", "level", "page_number", "dest", "struct_elem")
    name: str
    level: int
    page_number: int
    dest: DestinationXYZ
    struct_elem: Optional[StructElem]

    def __init__(
        self,
        name: str,
        level: int,
        page_number: int,
        dest: DestinationXYZ,
        struct_elem: Optional[StructElem] = None,
    ) -> None:
        self.name = name
        self.level = level
        self.page_number = page_number
        self.dest = dest
        self.struct_elem = struct_elem


class OutlineItemDictionary(PDFObject):
    __slots__ = (  # RAM usage optimization
        "_id",
        "title",
        "parent",
        "prev",
        "next",
        "first",
        "last",
        "count",
        "dest",
        "struct_elem",
    )

    def __init__(
        self,
        title: str,
        dest: Optional[DestinationXYZ] = None,
        struct_elem: Optional[StructElem] = None,
    ) -> None:
        super().__init__()
        self.title = PDFString(title, encrypt=True)
        self.parent: Optional[OutlineDictionary | OutlineItemDictionary] = None
        self.prev: Optional[OutlineItemDictionary] = None
        self.next: Optional[OutlineItemDictionary] = None
        self.first: Optional[OutlineItemDictionary] = None
        self.last: Optional[OutlineItemDictionary] = None
        self.count: int = 0
        self.dest: Optional[DestinationXYZ] = dest
        self.struct_elem: Optional[StructElem] = struct_elem

    def __str__(self) -> str:
        return f"OutlineItemDictionary(title={self.title}, dest={self.dest})"


class OutlineDictionary(PDFObject):
    __slots__ = ("_id", "type", "first", "last", "count")  # RAM usage optimization

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.type = "/Outlines"
        self.first = None
        self.last = None
        self.count = 0

    def __str__(self) -> str:
        return f"OutlineDictionary(count={self.count})"


def build_outline_objs(
    sections: Sequence[OutlineSection],
) -> Iterator[OutlineDictionary | OutlineItemDictionary]:
    """
    Build PDF objects constitutive of the documents outline,
    and yield them one by one, starting with the outline dictionary
    """
    outline = OutlineDictionary()
    yield outline

    outline_items: list[OutlineItemDictionary] = []
    last_outline_item_per_level: dict[int, OutlineItemDictionary] = {}
    parent_outline_item: OutlineDictionary | OutlineItemDictionary

    for section in sections:
        outline_item = OutlineItemDictionary(
            title=section.name,
            dest=section.dest,
            struct_elem=section.struct_elem,
        )
        yield outline_item

        if section.level in last_outline_item_per_level:
            last_outline_item_at_level = last_outline_item_per_level[section.level]
            last_outline_item_at_level.next = outline_item
            outline_item.prev = last_outline_item_at_level

        if section.level - 1 in last_outline_item_per_level:
            parent_outline_item = last_outline_item_per_level[section.level - 1]
        else:
            parent_outline_item = outline

        outline_item.parent = parent_outline_item
        if parent_outline_item.first is None:
            parent_outline_item.first = (  # pyright: ignore[reportAttributeAccessIssue]
                outline_item
            )
        parent_outline_item.last = (  # pyright: ignore[reportAttributeAccessIssue]
            outline_item
        )
        parent_outline_item.count += 1

        outline_items.append(outline_item)
        last_outline_item_per_level[section.level] = outline_item
        last_outline_item_per_level = {
            level: oitem
            for level, oitem in last_outline_item_per_level.items()
            if level <= section.level
        }


class TableOfContents:
    """
    A reference implementation of a Table of Contents (ToC) for use with `fpdf2`.

    This class provides a customizable Table of Contents that can be used directly or subclassed
    for additional functionality. To use this class, create an instance of `TableOfContents`,
    configure it as needed, and pass its `render_toc` method as the `render_toc_function` argument
    to `FPDF.insert_toc_placeholder()`.
    """

    def __init__(
        self,
        text_style: Optional[TextStyle] = None,
        use_section_title_styles: bool = False,
        level_indent: float = 7.5,
        line_spacing: float = 1.5,
        ignore_pages_before_toc: bool = True,
    ) -> None:
        self.text_style = text_style or TextStyle()
        self.use_section_title_styles = use_section_title_styles
        self.level_indent = level_indent
        self.line_spacing = line_spacing
        self.ignore_pages_before_toc = ignore_pages_before_toc

    def get_text_style(self, pdf: "FPDF", item: OutlineSection) -> TextStyle:
        if self.use_section_title_styles and pdf.section_title_styles[item.level]:
            return pdf.section_title_styles[item.level]
        if isinstance(self.text_style.l_margin, (str, Align)):
            raise ValueError(
                f"Unsupported l_margin value provided as TextStyle: {self.text_style.l_margin}"
            )
        return self.text_style

    def render_toc_item(self, pdf: "FPDF", item: OutlineSection) -> None:
        link = pdf.add_link(page=item.page_number)
        page_label = pdf.pages[item.page_number].get_label()

        # render the text on the left
        with pdf.use_text_style(self.get_text_style(pdf, item)):
            indent = item.level * self.level_indent
            pdf.set_x(pdf.l_margin + indent)
            pdf.multi_cell(
                w=pdf.epw - indent,
                text=item.name,
                new_x=XPos.END,
                new_y=YPos.LAST,
                link=link,
                align=Align.J,
                h=pdf.font_size * self.line_spacing,
            )

            # fill in-between with dots
            clearance_margin = pdf.c_margin * 2
            current_x = pdf.get_x()
            page_label_length = pdf.get_string_width(page_label)
            in_between_space = (
                pdf.w - current_x - page_label_length - clearance_margin - pdf.r_margin
            )
            if in_between_space < 0:
                # no space to render the page number - go to next line
                pdf.ln()
                pdf.set_x(pdf.l_margin + indent)
                current_x = pdf.get_x()
                in_between_space = pdf.w - current_x - page_label_length - pdf.r_margin
            in_between = ""
            if in_between_space > 0:
                while FloatTolerance.less_than(
                    pdf.get_string_width(in_between) + clearance_margin,
                    in_between_space,
                ):
                    in_between += "."

                if len(in_between) > 1:
                    pdf.multi_cell(
                        w=pdf.w - current_x - pdf.r_margin,
                        text=in_between[:-1],
                        new_x=XPos.END,
                        new_y=YPos.LAST,
                        link=link,
                        align=Align.L,
                        h=pdf.font_size * self.line_spacing,
                    )

            # render the page number on the right
            pdf.set_x(current_x)
            pdf.multi_cell(
                w=pdf.w - current_x - pdf.r_margin,
                text=page_label,
                new_x=XPos.LMARGIN,
                new_y=YPos.NEXT,
                link=link,
                align=Align.R,
                h=pdf.font_size * self.line_spacing,
            )

    def render_toc(self, pdf: "FPDF", outline: list[OutlineSection]) -> None:
        "This method can be overridden by subclasses to customize the Table of Contents style."
        assert pdf.toc_placeholder is not None
        for section in outline:
            if (
                self.ignore_pages_before_toc
                and section.page_number <= pdf.toc_placeholder.start_page
            ):
                continue
            self.render_toc_item(pdf, section)



from pathlib import Path

import fpdf
from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent


def test_alias_nb_pages(tmp_path):
    pdf = fpdf.FPDF()
    pdf.set_font("Times")
    pdf.add_page()
    pdf.cell(0, 10, f"Page {pdf.page_no()}/{{nb}}", align="C")
    pdf.add_page()
    pdf.cell(0, 10, f"Page {pdf.page_no()}/{{nb}}", align="C")
    assert_pdf_equal(pdf, HERE / "alias_nb_pages.pdf", tmp_path)


def test_custom_alias_nb_pages(tmp_path):
    pdf = fpdf.FPDF()
    pdf.set_font("Times")
    alias = "n{}b"
    # Prerequisite to get exactly the same output in the PDF:
    # the default alias and the new one must be of same width:
    assert pdf.get_string_width(pdf.str_alias_nb_pages) == pdf.get_string_width(alias)
    pdf.alias_nb_pages(alias)
    pdf.add_page()
    pdf.cell(0, 10, f"Page {pdf.page_no()}/{alias}", align="C")
    pdf.add_page()
    pdf.cell(0, 10, f"Page {pdf.page_no()}/{alias}", align="C")
    assert_pdf_equal(pdf, HERE / "alias_nb_pages.pdf", tmp_path)


def test_page_label(tmp_path):
    pdf = fpdf.FPDF()

    # title page - no numbering
    pdf.add_page(label_style="D")
    pdf.set_font("helvetica", "", 20)
    pdf.start_section("Title")
    pdf.cell(w=pdf.epw, text="TITLE", align="C")

    # pre-textual elements - lowercase roman numbering starting from 1
    pdf.add_page(label_style="r", label_start=1)
    pdf.set_font("helvetica", "", 14)
    pdf.start_section("Abstract")
    pdf.cell(text="Abstract")

    pdf.add_page()
    pdf.start_section("Table of contents")
    pdf.cell(text="Table of contents")

    pdf.add_page()
    pdf.start_section("List of figures and tables")
    pdf.cell(text="List of figures and tables")

    pdf.add_page()
    pdf.start_section("List of abbreviations")
    pdf.cell(text="List of abbreviations")

    pdf.add_page()
    pdf.start_section("Glossary")
    pdf.cell(text="Glossary")

    # textual elements - arabic numbers starting from 1
    pdf.add_page(label_style="D", label_start=1)
    pdf.set_font("helvetica", "", 14)
    pdf.start_section("Introduction")
    pdf.cell(text="Introduction")

    pdf.add_page()
    pdf.start_section("Literature review")
    pdf.cell(text="Literature review")

    pdf.add_page()
    pdf.start_section("Methodology")
    pdf.cell(text="Methodology")

    pdf.add_page()
    pdf.start_section("Results")
    pdf.cell(text="Results")

    pdf.add_page()
    pdf.start_section("Discussion")
    pdf.cell(text="Discussion")

    pdf.add_page()
    pdf.start_section("Conclusion")
    pdf.cell(text="Conclusion")

    pdf.add_page()
    pdf.start_section("Reference list")
    pdf.cell(text="Reference list")

    # appendices - prefix "A-" starting from 1
    pdf.add_page(label_style="D", label_prefix="A-", label_start=1)
    pdf.start_section("Appendices")
    pdf.cell(text="Appendix 1")

    pdf.add_page()
    pdf.cell(text="Appendix 2")

    pdf.add_page()
    pdf.cell(text="Appendix 3")

    pdf.add_page()
    pdf.cell(text="Appendix 4")

    pdf.add_page()
    pdf.cell(text="Appendix 5")

    assert_pdf_equal(pdf, HERE / "page_label.pdf", tmp_path)


def test_alias_with_shaping(tmp_path):
    pdf = fpdf.FPDF()
    pdf.add_font("Quicksand", style="", fname=HERE / "fonts" / "Quicksand-Regular.otf")
    pdf.add_page()
    pdf.set_font("Quicksand", size=24)
    pdf.set_text_shaping(True)
    pdf.write(text="Pages {nb}")
    pdf.ln()
    pdf.cell(text="{nb}", new_x="left", new_y="next")
    pdf.write_html("<h1>{nb}</h1>")
    pdf.multi_cell(w=pdf.epw, text="Number of pages: {nb}\nAgain:{nb}")
    pdf.add_page()
    pdf.set_text_shaping(False)
    pdf.write(text="Pages {nb}")
    pdf.ln()
    pdf.cell(text="{nb}", new_x="left", new_y="next")
    pdf.write_html("<h1>{nb}</h1>")
    pdf.multi_cell(w=pdf.epw, text="Number of pages: {nb}\nAgain:{nb}")
    assert_pdf_equal(pdf, HERE / "alias_with_text_shaping.pdf", tmp_path)



#!/usr/bin/env python3
"""Speed benchmark: how much time each lib takes to generate a 10 thousands pages PDF with ~180 distinct images?
(disclaimer: the author of this benchmark is fpdf2 current maintainer)"""

# Script used in this article: https://chezsoi.org/lucas/blog/fpdf2-5-2-svg-support-and-borb.html
import resource
from decimal import Decimal
from os.path import getsize
from pathlib import Path
from time import perf_counter

from fpdf import FPDF, __version__ as fpdf2_version
from borb.io.write import ascii_art
from borb.pdf.canvas.layout.image.image import Image
from borb.pdf.canvas.layout.page_layout.multi_column_layout import MultiColumnLayout
from borb.pdf.document.document import Document
from borb.pdf.page.page import Page
from borb.pdf.pdf import PDF
from PIL import Image as PILImage


class CustomMultiColumnLayoutWith(MultiColumnLayout):
    """
    Custom subclass allowing us to specify a custom inter-column margin width.
    Suggested as a feature in this PR: https://github.com/jorisschellekens/borb/pull/96
    """

    def __init__(self, *args, gutter_width, **kwargs):
        super().__init__(*args, **kwargs)
        if gutter_width is None:
            gutter_width = 0.05
        self._inter_column_margin = self._page_width * Decimal(gutter_width)
        self._column_width = (
            self._page_width
            - Decimal(2) * self._horizontal_margin
            - (self._number_of_columns - 1) * self._inter_column_margin
        ) / self._number_of_columns


PAGES_COUNT = 1000
HERE = Path(__file__).resolve().parent
if not (HERE / "../test/image/").is_dir():
    raise EnvironmentError(
        "This script depends on PNG files present in fpdf2 repository: it cannot be executed as standalone"
    )
PNG_FILE_PATHS = []
for img_path in (HERE / "../test/image/png_images/").glob("*.png"):
    # We skip the images that make borb crashes:
    if img_path.name != "51a4d21670dc8dfa8ffc9e54afd62f5f.png":
        PNG_FILE_PATHS.append(img_path)
for img_path in (HERE / "../test/image/png_test_suite/").glob("*.png"):
    # We skip the images that make borb crashes:
    if not img_path.name.startswith("x") and not img_path.name.endswith("4a08.png"):
        PNG_FILE_PATHS.append(img_path)


def fpdf2_intense_image_rendering(images):
    pdf = FPDF()
    pdf.set_margin(0)
    width, height = pdf.epw / 13, pdf.eph / 14
    for _ in range(PAGES_COUNT):
        pdf.add_page()
        for i, image in enumerate(images):
            x = (i // 14) * width
            y = (i % 14) * height
            pdf.image(image, x=x, y=y, w=width, h=height)

    memory_peak = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
    print(f"Memory usage peak (resource.ru_maxrss): {memory_peak // 1000}MB")
    pdf.output("fpdf2-10000-pages.pdf")
    print(
        f"Generated PDF file size: {getsize('fpdf2-10000-pages.pdf')/1000/1000:.2f}MB"
    )


def borb_intense_image_rendering(images):
    doc = Document()
    width, height = Decimal(595) / 13, Decimal(842) / 14
    for _ in range(PAGES_COUNT):
        page = Page()
        doc.append_page(page)
        layout = CustomMultiColumnLayoutWith(
            page,
            number_of_columns=13,
            gutter_width=0,
            horizontal_margin=0,
            vertical_margin=0,
        )
        for _, image in enumerate(images):
            layout.add(
                Image(
                    image,
                    width=width,
                    height=height,
                    # Zero-margins are currently ignored, resulting in page breaks: https://github.com/jorisschellekens/borb/pull/95
                    margin_bottom=0,
                    margin_left=0,
                    margin_right=0,
                    margin_top=0,
                )
            )

    memory_peak = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
    print(f"Memory usage peak (resource.ru_maxrss): {memory_peak // 1000}MB")
    with open("borb-10000-pages.pdf", "wb") as pdf_file:
        PDF.dumps(pdf_file, doc)
    print(f"Generated PDF file size: {getsize('borb-10000-pages.pdf')/1000/1000:.2f}MB")


if __name__ == "__main__":
    print(__doc__)

    ascii_logo_path = Path(ascii_art.__file__).resolve().parent / "ascii_logo.txt"
    with ascii_logo_path.open(encoding="utf8") as text_file:
        borb_version = text_file.readline().strip()
    print(f"Versions tested: {borb_version} VS fpdf2 v{fpdf2_version}")

    print("\nBenchmarking fpdf2...")
    start = perf_counter()
    fpdf2_intense_image_rendering(PNG_FILE_PATHS)
    duration = perf_counter() - start
    print(f"Duration: {duration:.2f}s")

    print("\nBenchmarking borb...")
    start = perf_counter()
    # borb_intense_image_rendering(PNG_FILE_PATHS)  # This crashes:
    # We preload Pillow images to avoid borb raising an OSError: [Errno 24] Too many open files
    pillow_images = [PILImage.open(png_file_path) for png_file_path in PNG_FILE_PATHS]
    borb_intense_image_rendering(pillow_images)
    duration = perf_counter() - start
    print(f"Duration: {duration:.2f}s")



from pathlib import Path

import fpdf
from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent


def next_row(pdf):
    pdf.ln()
    pdf.set_y(pdf.get_y() + size + margin)


size = 50
margin = 10


def test_ellipse_not_circle(tmp_path):
    pdf = fpdf.FPDF(unit="mm")
    pdf.add_page()

    for counter, style in enumerate(["", "F", "FD", "DF", None]):
        pdf.ellipse(x=pdf.get_x(), y=pdf.get_y(), w=size / 2, h=size, style=style)
        pdf.set_x(pdf.get_x() + (size / 2) + margin)
        if counter % 3 == 0:
            next_row(pdf)

    assert_pdf_equal(pdf, HERE / "class_ellipse_not_circle.pdf", tmp_path)


def test_ellipse_style(tmp_path):
    pdf = fpdf.FPDF(unit="mm")
    pdf.add_page()

    for counter, style in enumerate(["", "F", "FD", "DF", None]):
        pdf.ellipse(x=pdf.get_x(), y=pdf.get_y(), w=size, h=size, style=style)
        pdf.set_x(pdf.get_x() + size + margin)
        if counter % 3 == 0:
            next_row(pdf)

    assert_pdf_equal(pdf, HERE / "class_ellipse_style.pdf", tmp_path)


def test_ellipse_line_width(tmp_path):
    pdf = fpdf.FPDF(unit="mm")
    pdf.add_page()

    for line_width in [1, 2, 3]:
        pdf.set_line_width(line_width)
        pdf.ellipse(x=pdf.get_x(), y=pdf.get_y(), w=size, h=size, style=None)
        pdf.set_x(pdf.get_x() + size + margin)
    next_row(pdf)
    for line_width in [4, 5, 6]:
        pdf.set_line_width(line_width)
        pdf.ellipse(x=pdf.get_x(), y=pdf.get_y(), w=size, h=size, style=None)
        pdf.set_x(pdf.get_x() + size + margin)
    pdf.set_line_width(0.2)  # reset

    assert_pdf_equal(pdf, HERE / "class_ellipse_line_width.pdf", tmp_path)


def test_ellipse_draw_color(tmp_path):
    pdf = fpdf.FPDF(unit="mm")
    pdf.add_page()

    pdf.set_line_width(0.5)
    for gray in [70, 140, 210]:
        pdf.set_draw_color(gray)
        pdf.ellipse(x=pdf.get_x(), y=pdf.get_y(), w=size, h=size, style=None)
        pdf.set_x(pdf.get_x() + size + margin)

    assert_pdf_equal(pdf, HERE / "class_ellipse_draw_color.pdf", tmp_path)


def test_ellipse_fill_color(tmp_path):
    pdf = fpdf.FPDF(unit="mm")
    pdf.add_page()

    pdf.set_fill_color(240)
    for color in [[230, 30, 180], [30, 180, 30], [30, 30, 70]]:
        pdf.set_draw_color(*color)
        pdf.ellipse(x=pdf.get_x(), y=pdf.get_y(), w=size, h=size, style="FD")
        pdf.set_x(pdf.get_x() + size + margin)
    next_row(pdf)

    assert_pdf_equal(pdf, HERE / "class_ellipse_fill_color.pdf", tmp_path)



from pathlib import Path
from fpdf import FPDF
from fpdf.pattern import LinearGradient
from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent


def test_linear_gradient_extend(tmp_path):
    pdf = FPDF()

    pdf.add_page()

    pdf.set_font("helvetica", "", 20)
    pdf.cell(
        text="Creating a gradient slightly smaller than the rectangle",
        new_x="LEFT",
        new_y="NEXT",
    )
    x = pdf.l_margin
    y = pdf.get_y()
    with pdf.use_pattern(
        LinearGradient(
            pdf.l_margin + 10,
            0,
            pdf.epw + pdf.l_margin - 10,
            0,
            ["#C33764", "#1D2671"],
        )
    ):
        pdf.rect(x=x, y=y, w=pdf.epw, h=20, style="FD")
        y += 25
        pdf.set_y(y)
        pdf.set_font("helvetica", "", 40)
        pdf.cell(
            text="LINEAR GRADIENT", align="C", w=pdf.epw, new_x="LEFT", new_y="NEXT"
        )

    pdf.ln()
    pdf.set_font("helvetica", "", 20)
    pdf.cell(text="Adding a background color", new_x="LEFT", new_y="NEXT")
    x = pdf.l_margin
    y = pdf.get_y()
    with pdf.use_pattern(
        LinearGradient(
            pdf.l_margin + 10,
            0,
            pdf.epw + pdf.l_margin - 10,
            0,
            ["#C33764", "#1D2671"],
            background="#868F96",
        )
    ):
        pdf.rect(x=x, y=y, w=pdf.epw, h=20, style="FD")
        y += 25
        pdf.set_y(y)
        pdf.set_font("helvetica", "", 40)
        pdf.cell(
            text="LINEAR GRADIENT", align="C", w=pdf.epw, new_x="LEFT", new_y="NEXT"
        )

    pdf.ln()
    pdf.set_font("helvetica", "", 20)
    pdf.cell(text="Adding extend before", new_x="LEFT", new_y="NEXT")
    x = pdf.l_margin
    y = pdf.get_y()
    with pdf.use_pattern(
        LinearGradient(
            pdf.l_margin + 10,
            0,
            pdf.epw + pdf.l_margin - 10,
            0,
            ["#C33764", "#1D2671"],
            background="#868F96",
            extend_before=True,
        )
    ):
        pdf.rect(x=x, y=y, w=pdf.epw, h=20, style="FD")
        y += 25
        pdf.set_y(y)
        pdf.set_font("helvetica", "", 40)
        pdf.cell(
            text="LINEAR GRADIENT", align="C", w=pdf.epw, new_x="LEFT", new_y="NEXT"
        )

    pdf.ln()
    pdf.set_font("helvetica", "", 20)
    pdf.cell(text="Adding extend after", new_x="LEFT", new_y="NEXT")
    x = pdf.l_margin
    y = pdf.get_y()
    with pdf.use_pattern(
        LinearGradient(
            pdf.l_margin + 10,
            0,
            pdf.epw + pdf.l_margin - 10,
            0,
            ["#C33764", "#1D2671"],
            background="#868F96",
            extend_before=True,
            extend_after=True,
        )
    ):
        pdf.rect(x=x, y=y, w=pdf.epw, h=20, style="FD")
        y += 25
        pdf.set_y(y)
        pdf.set_font("helvetica", "", 40)
        pdf.cell(
            text="LINEAR GRADIENT", align="C", w=pdf.epw, new_x="LEFT", new_y="NEXT"
        )

    assert_pdf_equal(
        pdf,
        HERE / "linear_gradient_extend.pdf",
        tmp_path,
    )


def test_linear_gradient_multiple_colors(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    x = pdf.l_margin
    y = pdf.get_y()
    with pdf.use_pattern(
        LinearGradient(
            pdf.l_margin,
            0,
            pdf.epw + pdf.l_margin,
            0,
            ["#868F96", "#596164", "#537895", "#09203F"],
        )
    ):
        pdf.rect(x=x, y=y, w=pdf.epw, h=20, style="FD")
    y += 25

    with pdf.use_pattern(
        LinearGradient(
            pdf.l_margin,
            0,
            pdf.epw + pdf.l_margin,
            0,
            ["#FFECD2", "#FCB69F", "#DD2476"],
        )
    ):
        pdf.rect(x=x, y=y, w=pdf.epw, h=20, style="FD")

    assert_pdf_equal(
        pdf,
        HERE / "linear_gradient_multiple_colors.pdf",
        tmp_path,
    )


def test_linear_gradient_vertical(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    x = pdf.l_margin
    y = pdf.get_y()
    with pdf.use_pattern(
        LinearGradient(
            0,
            y,
            0,
            y + 50,
            ["#92EFFD", "#4E65FF"],
        )
    ):
        pdf.rect(x=x, y=y, w=pdf.epw, h=50, style="FD")
    y += 55

    with pdf.use_pattern(
        LinearGradient(
            0,
            y,
            0,
            y + 50,
            ["#FFECD2", "#FCB69F", "#DD2476"],
        )
    ):
        pdf.rect(x=x, y=y, w=pdf.epw, h=50, style="FD")

    assert_pdf_equal(pdf, HERE / "linear_gradient_vertical.pdf", tmp_path)


def test_linear_gradient_diagonal(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    x = pdf.l_margin
    y = pdf.get_y()
    with pdf.use_pattern(
        LinearGradient(
            pdf.l_margin,
            y,
            pdf.epw + pdf.l_margin,
            y + 100,
            ["#92EFFD", "#4E65FF"],
        )
    ):
        pdf.rect(x=x, y=y, w=pdf.epw, h=100, style="FD")
    y += 105

    with pdf.use_pattern(
        LinearGradient(
            pdf.l_margin,
            y + 100,
            pdf.epw + pdf.l_margin,
            y,
            ["#FFECD2", "#DD2476", "#FCB69F"],
        )
    ):
        pdf.rect(x=x, y=y, w=pdf.epw, h=100, style="FD")

    assert_pdf_equal(
        pdf,
        HERE / "linear_gradient_diagonal.pdf",
        tmp_path,
    )



from pathlib import Path
import warnings

from fpdf import FPDF, HTML2FPDF
from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent


def test_html_toc(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.write_html("""
        <h1>Document title</h1>
        <br><br><br>
        <u>Table of content:</u>
        <br>
        <toc></toc>
        <section><h2>Subtitle 1</h2>
            <section><h3>Subtitle 1.1</h3>
            Lorem ipsum dolor sit amet, consectetur adipiscing elit,
            sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
            </section>
            <br><br><br><br><br><br><br><br><br><br><br><br><br><br><br><br><br><br><br><br>
            <section><h3>Subtitle 1.2</h3>
            Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat.
            </section>
        </section>
        <br><br><br><br><br><br><br><br><br><br><br><br><br><br><br><br><br><br><br><br>
        <br><br><br><br><br><br><br><br><br><br><br><br><br><br><br><br><br><br><br><br>
        <section><h2>Subtitle 2</h2><br>
            <section><h3>Subtitle 2.1</h3>
            Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur.
            </section>
            <br><br><br><br><br><br><br><br><br><br><br><br><br><br><br><br><br><br><br><br>
            <section><h3>Subtitle 2.2</h3>
            Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.
            </section>
        </section>
        """)
    assert_pdf_equal(pdf, HERE / "html_toc.pdf", tmp_path)


def test_html_toc_2_pages(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.write_html("""<h1>Document title</h1>
        <br><br><br>
        <u>Table of content:</u>
        <br>
        <toc pages="2"></toc>
        <h2>Subtitle 0</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 1</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 2</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 3</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 4</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 5</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 6</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 7</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 8</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 9</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 10</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 11</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 12</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 13</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 14</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 15</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 16</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 17</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 18</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 19</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 20</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 21</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 22</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 23</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 24</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 25</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 26</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 27</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 28</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 29</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 30</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 31</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 32</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 33</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 34</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 35</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 36</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 37</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 38</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 39</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 40</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 41</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 42</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 43</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 44</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 45</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 46</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 47</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 48</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 49</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 50</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 51</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 52</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 53</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 54</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        <h2>Subtitle 55</h2>
        Lorem ipsum dolor sit amet, consectetur adipiscing elit,
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.
        """)
    assert_pdf_equal(pdf, HERE / "html_toc_2_pages.pdf", tmp_path)


def test_html_toc_with_h1_as_2nd_heading(tmp_path):  # issue 239
    pdf = FPDF()
    pdf.add_page()
    pdf.write_html("""<toc></toc>
        <h2>?-1</h2>
        <h3>?-1-1</h3>
        <h1>1</h1>
        <h2>1-1</h2>
        <h3>1-1-1</h3>""")
    assert_pdf_equal(pdf, HERE / "html_toc_with_h1_as_2nd_heading.pdf", tmp_path)


class CustomHTML2FPDF(HTML2FPDF):
    def render_toc(self, pdf, outline):
        pdf.cell(text="Table of contents:", new_x="LMARGIN", new_y="NEXT")
        for section in outline:
            pdf.cell(
                text=f"* {section.name} (page {section.page_number})",
                new_x="LMARGIN",
                new_y="NEXT",
            )


def test_html_toc_with_custom_rendering(tmp_path):  # issue 240 & 670
    class PDF(FPDF):
        HTML2FPDF_CLASS = CustomHTML2FPDF

    pdf = PDF()
    pdf.add_page()

    # Ensure no warning is raised:
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        pdf.write_html("""<toc></toc>
        <h1>Level 1</h1>
        <h2>Level 2</h2>
        <h3>Level 3</h3>
        <h4>Level 4</h4>
        <h5>Level 5</h5>
        <h6>Level 6</h6>
        <p>paragraph</p>""")
    assert_pdf_equal(pdf, HERE / "html_toc_with_custom_rendering.pdf", tmp_path)



"""
An example of script that generates a one page invoice with barcode,
with data coming from a CSV file.
"""

import os

from fpdf import FPDF


class Form:
    def __init__(self, infile):
        keys = (
            "name",
            "type",
            "x1",
            "y1",
            "x2",
            "y2",
            "font",
            "size",
            "bold",
            "italic",
            "underline",
            "foreground",
            "background",
            "align",
            "text",
            "priority",
        )
        # parse form format file and create fields dict
        self.fields = {}
        with open(infile, encoding="utf8") as file:
            for linea in file.readlines():
                kargs = {}
                for i, v in enumerate(linea.split(";")):
                    if v in ("", "None"):
                        v = None
                    elif v.startswith("'"):
                        v = v.strip()[1:-1]
                    else:
                        v = float(v.replace(",", "."))
                    kargs[keys[i]] = v
                self.fields[kargs["name"].lower()] = kargs
        self.handlers = {
            "T": self.text,
            "L": self.line,
            "I": self.image,
            "B": self.rect,
            "BC": self.barcode,
        }

    def set(self, name, value):
        if name.lower() in self.fields:
            self.fields[name.lower()]["text"] = value

    def render(self, outfile):
        pdf = FPDF()
        pdf.add_page()
        pdf.set_font("helvetica", style="B", size=16)

        for field in self.fields.values():
            self.handlers[field["type"].upper()](pdf, **field)

        pdf.output(outfile)

    @staticmethod
    def text(
        pdf,
        *_,
        x1=0,
        y1=0,
        x2=0,
        y2=0,
        text="",
        font="helvetica",
        size=10,
        bold=False,
        italic=False,
        underline=False,
        align="",
        **__,
    ):
        if text:
            font = font.strip().lower()
            style = ""
            if bold:
                style += "B"
            if italic:
                style += "I"
            if underline:
                style += "U"
            align = {"I": "L", "D": "R", "C": "C", "": "", None: None}[align]
            pdf.set_font(font, style, size)
            # m_k = 72 / 2.54
            # h = (size/m_k)
            pdf.set_xy(x1, y1)
            pdf.cell(w=x2 - x1, h=y2 - y1, txt=text, align=align)
            # pdf.Text(x=x1,y=y1,txt=text)

    @staticmethod
    def line(pdf, *_, x1=0, y1=0, x2=0, y2=0, size=0, **__):
        pdf.set_line_width(size)
        pdf.line(x1, y1, x2, y2)

    @staticmethod
    def rect(pdf, *_, x1=0, y1=0, x2=0, y2=0, size=0, **__):
        pdf.set_line_width(size)
        pdf.rect(x1, y1, x2 - x1, y2 - y1)

    @staticmethod
    def image(pdf, *_, x1=0, y1=0, x2=0, y2=0, text="", **__):
        pdf.image(text, x1, y1, w=x2 - x1, h=y2 - y1)

    @staticmethod
    def barcode(pdf, *_, x1=0, y1=0, text="", font="helvetica", size=1, **__):
        font = font.lower().strip()
        if font == "interleaved 2of5 nt":
            pdf.interleaved2of5(text, x1, y1, w=size)


if __name__ == "__main__":
    os.chdir(os.path.dirname(os.path.realpath(__file__)))
    f = Form("invoice.csv")
    f.set("EMPRESA", "Saraza")
    f.set("logo", "logo.png")
    f.render("./invoice.pdf")



import csv
from fpdf import FPDF
from fpdf.fonts import FontFace
from fpdf.enums import TableCellFillMode

with open("countries.txt", encoding="utf8") as csv_file:
    data = list(csv.reader(csv_file, delimiter=","))

pdf = FPDF()
pdf.set_font("helvetica", size=14)

# Basic table:
pdf.add_page()
with pdf.table() as table:
    for data_row in data:
        row = table.row()
        for datum in data_row:
            row.cell(datum)

# Styled table:
pdf.add_page()
pdf.set_draw_color(255, 0, 0)
pdf.set_line_width(0.3)
headings_style = FontFace(emphasis="BOLD", color=255, fill_color=(255, 100, 0))
with pdf.table(
    borders_layout="NO_HORIZONTAL_LINES",
    cell_fill_color=(224, 235, 255),
    cell_fill_mode=TableCellFillMode.ROWS,
    col_widths=(42, 39, 35, 42),
    headings_style=headings_style,
    line_height=6,
    text_align=("LEFT", "CENTER", "RIGHT", "RIGHT"),
    width=160,
) as table:
    for data_row in data:
        row = table.row()
        for datum in data_row:
            row.cell(datum)

pdf.output("tuto5.pdf")



import socket
import pytest

from pathlib import Path
from urllib.error import HTTPError

import fpdf
from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent
PNG_IMG_URL = "https://placehold.co/300x200.png"


def test_png_url(tmp_path):
    pdf = fpdf.FPDF()
    pdf.add_page()
    try:
        pdf.image(PNG_IMG_URL, x=15, y=15, w=30, h=25)
    except (HTTPError, socket.timeout) as e:
        pytest.skip(f"Network error: {e}")
    assert_pdf_equal(pdf, HERE / "image_png_url.pdf", tmp_path)



from fpdf.util import escape_parens


def test_escape_parens_string():
    expected = "\\\\ \\) \\( \\r"
    assert expected == escape_parens("\\ ) ( \r")


def test_escape_parens_bytes():
    expected = "\\\\ \\) \\( \\r".encode()
    assert expected == escape_parens("\\ ) ( \r".encode())



from pathlib import Path

from fpdf import FPDF, TextMode

import pytest

from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent


def test_text_modes(tmp_path):
    pdf = FPDF(format=(350, 150))
    pdf.add_page()
    pdf.set_font("Helvetica", size=80)
    with pdf.local_context(fill_color=(255, 128, 0)):
        pdf.cell(text="FILL default")
    with pdf.local_context(text_color=(0, 128, 255)):
        pdf.cell(text=" text mode")
    pdf.ln()
    with pdf.local_context(text_mode=TextMode.STROKE, line_width=2):
        pdf.cell(text="STROKE text mode")
    pdf.ln()
    pdf.text_mode = TextMode.FILL_STROKE
    pdf.line_width = 4
    pdf.set_draw_color(255, 0, 255)
    pdf.cell(text="FILL_STROKE text mode")
    pdf.ln()
    with pdf.local_context():
        pdf.text_mode = "INVISIBLE"  # testing TextMode.coerce
        pdf.cell(text="INVISIBLE text mode")
    assert_pdf_equal(pdf, HERE / "text_modes.pdf", tmp_path)


def test_clip_text_modes(tmp_path):
    pdf = FPDF(format=(420, 150))
    pdf.add_page()
    pdf.set_font("Helvetica", size=80)
    pdf.line_width = 1
    with pdf.local_context(text_mode=TextMode.FILL_CLIP, text_color=(0, 255, 255)):
        pdf.cell(text="FILL_CLIP text mode")
        for r in range(0, 100, 1):
            pdf.circle(x=110, y=22, radius=r)
    pdf.ln()
    with pdf.local_context(text_mode=TextMode.STROKE_CLIP):
        pdf.cell(text="STROKE_CLIP text mode")
        for r in range(0, 100, 1):
            pdf.circle(x=110, y=50, radius=r)
    pdf.ln()
    with pdf.local_context(
        text_mode=TextMode.FILL_STROKE_CLIP, text_color=(0, 255, 255)
    ):
        pdf.cell(text="FILL_STROKE_CLIP text mode")
        for r in range(0, 100, 1):
            pdf.circle(x=110, y=78, radius=r)
    pdf.ln()
    with pdf.local_context(text_mode=TextMode.CLIP):
        pdf.cell(text="CLIP text mode")
        for r in range(0, 100, 1):
            pdf.circle(x=110, y=106, radius=r)
    pdf.ln()
    assert_pdf_equal(pdf, HERE / "clip_text_modes.pdf", tmp_path)


def test_invalid_text_mode():
    pdf = FPDF()
    with pytest.raises(ValueError):
        pdf.text_mode = "DUMMY"



from mistune import html

html = html("""
# Top title (ATX)

Subtitle (setext)
-----------------

### An even lower heading (ATX)

**Text in bold**

_Text in italics_

~~Strikethrough~~

[This is a link](https://github.com/PyFPDF/fpdf2)

<https://py-pdf.github.io/fpdf2/>

This is an unordered list:
* an item
* another item

This is an ordered list:
1. first item
2. second item
3. third item with an unordered sublist:
    * an item
    * another item

Inline `code span`

A table:

| Foo | Bar | Baz |
| ---:|:---:|:--- |
| Foo | Bar | Baz |

Actual HTML:

<dl>
  <dt>Term1</dt><dd>Definition1</dd>
  <dt>Term2</dt><dd>Definition2</dd>
</dl>

Some horizontal thematic breaks:

***
---
___

![Alternate description](https://py-pdf.github.io/fpdf2/fpdf2-logo.png)
""")

from fpdf import FPDF

pdf = FPDF()
pdf.add_page()
pdf.write_html(html)
pdf.output("pdf-from-markdown-with-mistune.pdf")



from pathlib import Path

from fpdf import FPDF
from test.conftest import assert_pdf_equal

HERE = Path(__file__).resolve().parent

XMP_METADATA = """<x:xmpmeta xmlns:x="adobe:ns:meta/" x:xmptk="fpdf2">
  <rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#">
    <rdf:Description rdf:about="">
      <dc:title xmlns:dc="http://purl.org/dc/elements/1.1/">
        <rdf:Alt>
          <rdf:li xml:lang="x-default">{title}</rdf:li>
        </rdf:Alt>
      </dc:title>
    </rdf:Description>
    <rdf:Description rdf:about="">
      <dc:description xmlns:dc="http://purl.org/dc/elements/1.1/">
        <rdf:Alt>
          <rdf:li xml:lang="x-default">{desc}</rdf:li>
        </rdf:Alt>
      </dc:description>
    </rdf:Description>
    <rdf:Description rdf:about="">
      <dc:creator xmlns:dc="http://purl.org/dc/elements/1.1/">
        <rdf:Seq>
          <rdf:li>{creator}</rdf:li>
        </rdf:Seq>
      </dc:creator>
    </rdf:Description>
    <rdf:Description xmlns:pdf="http://ns.adobe.com/pdf/1.3/" rdf:about="" pdf:Keywords="{keywords}"/>
    <rdf:Description xmlns:pdf="http://ns.adobe.com/pdf/1.3/" rdf:about="" pdf:Producer="py-pdf/fpdf2.X.Y"/>
    <rdf:Description xmlns:xmp="http://ns.adobe.com/xap/1.0/" rdf:about="" xmp:CreatorTool="fpdf2"/>
  </rdf:RDF>
</x:xmpmeta>"""


def document_operations(doc):
    doc.add_page()
    doc.set_font("helvetica", size=12)
    doc.cell(w=72, h=0, border=1, text="hello world", fill=False, link="")


def test_put_info_all(tmp_path):
    """This test tests all possible inputs to FPDF#_put_info."""
    doc = FPDF()
    document_operations(doc)
    doc.set_title("sample title")
    doc.set_lang("en-US")
    doc.set_subject("sample subject")
    doc.set_author("sample author")
    doc.set_keywords("sample keywords")
    doc.set_creator("sample creator")
    doc.set_producer("py-pdf/fpdf2.X.Y")
    assert_pdf_equal(doc, HERE / "put_info_all.pdf", tmp_path)


def test_put_info_some(tmp_path):
    """This test tests some possible inputs to FPDF#_put_info."""
    doc = FPDF()
    document_operations(doc)
    doc.set_title("sample title")
    doc.set_keywords("sample keywords")
    doc.set_creator("sample creator")
    assert_pdf_equal(doc, HERE / "put_info_some.pdf", tmp_path)


def test_xmp_metadata(tmp_path):
    doc = FPDF()
    document_operations(doc)
    doc.set_xmp_metadata(
        XMP_METADATA.format(
            title="My document title",
            desc="This is a test document for fpdf2 with XMP metadata",
            creator="Lucas Cimon",
            keywords="test data pdf fpdf2",
        )
    )
    assert_pdf_equal(doc, HERE / "xmp_metadata.pdf", tmp_path)


def test_metadata_unicode(tmp_path):  # issue 685
    doc = FPDF()
    document_operations(doc)
    doc.set_creator("༄༅། །སྒྲུབ།")
    doc.set_title("༄༅། །སྒྲུབ།")
    doc.set_keywords("༄༅། །སྒྲུབ།")
    doc.set_creator("༄༅། །སྒྲུབ།")
    assert_pdf_equal(doc, HERE / "metadata_unicode.pdf", tmp_path)


def test_xmp_metadata_unicode(tmp_path):  # issue 685
    doc = FPDF()
    document_operations(doc)
    doc.set_xmp_metadata(
        XMP_METADATA.format(
            title="༄༅། །སྒྲུབ།",
            desc="༄༅། །སྒྲུབ།",
            creator="༄༅། །སྒྲུབ།",
            keywords="༄༅། །སྒྲུབ།",
        )
    )
    assert_pdf_equal(doc, HERE / "xmp_metadata_unicode.pdf", tmp_path)



import logging
from pathlib import Path

import pytest

from fpdf import FPDF, FPDFException
from fpdf.drawing import DeviceRGB
from fpdf.enums import (
    TableBordersLayout,
    TableBorderStyle,
    TableCellFillMode,
    TableCellStyle,
    TextEmphasis,
)
from fpdf.fonts import FontFace
from test.conftest import LOREM_IPSUM, assert_pdf_equal

HERE = Path(__file__).resolve().parent

TABLE_DATA = (
    ("First name", "Last name", "Age", "City"),
    ("Jules", "Smith", "34", "San Juan"),
    ("Mary", "Ramos", "45", "Orlando"),
    ("Carlson", "Banks", "19", "Los Angeles"),
    ("Lucas", "Cimon", "31", "Angers"),
)

MULTILINE_TABLE_DATA = (
    ("Extract", "Text length"),
    (LOREM_IPSUM[:200], str(len(LOREM_IPSUM[:200]))),
    (LOREM_IPSUM[200:400], str(len(LOREM_IPSUM[200:400]))),
    (LOREM_IPSUM[400:600], str(len(LOREM_IPSUM[400:600]))),
    (LOREM_IPSUM[600:800], str(len(LOREM_IPSUM[600:800]))),
    (LOREM_IPSUM[800:1000], str(len(LOREM_IPSUM[800:1000]))),
    (LOREM_IPSUM[1000:1200], str(len(LOREM_IPSUM[1000:1200]))),
)

MULTI_HEADING_TABLE_DATA = (
    ("Fruits", "Dairy"),
    ("Apple", "Banana", "Cherry", "Cheese", "Milk", "Yogurt"),
    ("1", "2", "3", "4", "5", "6"),
    ("2", "3", "4", "5", "6", "7"),
    ("3", "4", "5", "6", "7", "8"),
)


def test_table_simple(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    with pdf.table() as table:
        for data_row in TABLE_DATA:
            row = table.row()
            for datum in data_row:
                row.cell(datum)
    assert_pdf_equal(pdf, HERE / "table_simple.pdf", tmp_path)


def test_table_with_no_row():
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    with pdf.table():
        pass


def test_table_with_no_column():
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    with pdf.table() as table:
        for _ in TABLE_DATA:
            table.row()


def test_table_with_syntactic_sugar(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    with pdf.table(TABLE_DATA):
        pass
    assert_pdf_equal(pdf, HERE / "table_simple.pdf", tmp_path)

    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    with pdf.table() as table:
        table.row(TABLE_DATA[0])
        table.row(TABLE_DATA[1])
        table.row(TABLE_DATA[2])
        table.row(TABLE_DATA[3])
        table.row(TABLE_DATA[4])
    assert_pdf_equal(pdf, HERE / "table_simple.pdf", tmp_path)


def test_table_with_fixed_col_width(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    with pdf.table(TABLE_DATA, col_widths=pdf.epw / 5):
        pass
    assert_pdf_equal(pdf, HERE / "table_with_fixed_col_width.pdf", tmp_path)


def test_table_with_fixed_col_width_and_align(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    with pdf.table(TABLE_DATA, col_widths=pdf.epw / 5, align="L"):
        pass
    with pdf.table(TABLE_DATA, col_widths=pdf.epw / 5, align="C"):
        pass
    with pdf.table(TABLE_DATA, col_widths=pdf.epw / 5, align="R"):
        pass
    pdf.add_page()
    with pdf.table(TABLE_DATA, width=150, col_widths=(1, 1, 1, 1), align="L"):
        pass
    with pdf.table(TABLE_DATA, width=150, col_widths=(1, 1, 1, 1), align="C"):
        pass
    with pdf.table(TABLE_DATA, width=150, col_widths=(1, 1, 1, 1), align="R"):
        pass
    pdf.add_page()
    with pdf.table(TABLE_DATA, col_widths=(1, 1, 1, 1), align="L"):
        pass
    with pdf.table(TABLE_DATA, col_widths=(1, 1, 1, 1), align="C"):
        pass
    with pdf.table(TABLE_DATA, col_widths=(1, 1, 1, 1), align="R"):
        pass
    pdf.add_page()
    with pdf.table(TABLE_DATA, width=150, col_widths=(1, 1, 1, 1), align="L"):
        pass
    with pdf.table(TABLE_DATA, width=150, col_widths=(1, 1, 1, 1), align="C"):
        pass
    with pdf.table(TABLE_DATA, width=150, col_widths=(1, 1, 1, 1), align="R"):
        pass
    pdf.add_page()
    with pdf.table(TABLE_DATA, width=150, col_widths=37.5, align="L"):
        pass
    with pdf.table(TABLE_DATA, width=150, col_widths=37.5, align="C"):
        pass
    with pdf.table(TABLE_DATA, width=150, col_widths=37.5, align="R"):
        pass
    assert_pdf_equal(pdf, HERE / "table_with_fixed_col_width_and_align.pdf", tmp_path)


def test_table_with_varying_col_widths(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    with pdf.table(TABLE_DATA, col_widths=(30, 30, 10, 30)):
        pass
    assert_pdf_equal(pdf, HERE / "table_with_varying_col_widths.pdf", tmp_path)


def test_table_with_invalid_col_widths():
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    with pytest.raises(ValueError):
        with pdf.table(TABLE_DATA, col_widths=(20, 30, 50)):
            pass


def test_table_with_fixed_row_height(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    with pdf.table(TABLE_DATA, line_height=2.5 * pdf.font_size):
        pass
    assert_pdf_equal(pdf, HERE / "table_with_fixed_row_height.pdf", tmp_path)


def test_table_with_multiline_cells(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    with pdf.table() as table:
        for data_row in MULTILINE_TABLE_DATA:
            row = table.row()
            for datum in data_row:
                row.cell(datum)
    assert pdf.pages_count == 2
    assert_pdf_equal(pdf, HERE / "table_with_multiline_cells.pdf", tmp_path)


def test_table_with_multiline_cells_and_fixed_row_height(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    with pdf.table(line_height=2.5 * pdf.font_size) as table:
        for data_row in MULTILINE_TABLE_DATA:
            row = table.row()
            for datum in data_row:
                row.cell(datum)
    assert pdf.pages_count == 2

    assert_pdf_equal(
        pdf, HERE / "table_with_multiline_cells_and_fixed_row_height.pdf", tmp_path
    )


def test_table_with_fixed_width(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    with pdf.table(TABLE_DATA, width=150):
        pass
    assert_pdf_equal(pdf, HERE / "table_with_fixed_width.pdf", tmp_path)


def test_table_with_invalid_width():
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    with pytest.raises(ValueError):
        with pdf.table(TABLE_DATA, width=200):
            pass


def test_table_without_headings(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    with pdf.table(TABLE_DATA, first_row_as_headings=False):
        pass
    assert_pdf_equal(pdf, HERE / "table_without_headings.pdf", tmp_path)


def test_table_with_multiline_cells_and_without_headings(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    with pdf.table(first_row_as_headings=False) as table:
        for data_row in MULTILINE_TABLE_DATA + MULTILINE_TABLE_DATA[1:]:
            row = table.row()
            for datum in data_row:
                row.cell(datum)
    assert pdf.pages_count == 4
    assert_pdf_equal(
        pdf,
        HERE / "table_with_multiline_cells_and_without_headings.pdf",
        tmp_path,
    )


def test_table_with_headings_styled(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    blue = DeviceRGB(r=0, g=0, b=1)
    grey = 128
    headings_style = FontFace(
        emphasis=TextEmphasis.I | TextEmphasis.U, color=blue, fill_color=grey
    )
    with pdf.table(TABLE_DATA, headings_style=headings_style):
        pass
    assert_pdf_equal(pdf, HERE / "table_with_headings_styled.pdf", tmp_path)


def test_table_with_multiline_cells_and_split_over_3_pages(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    with pdf.table() as table:
        for data_row in MULTILINE_TABLE_DATA + MULTILINE_TABLE_DATA[1:]:
            row = table.row()
            for datum in data_row:
                row.cell(datum)
    assert pdf.pages_count == 4
    assert_pdf_equal(
        pdf, HERE / "table_with_multiline_cells_and_split_over_3_pages.pdf", tmp_path
    )


def test_table_with_cell_fill(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    greyscale = 200
    with pdf.table(TABLE_DATA, cell_fill_color=greyscale, cell_fill_mode="ROWS"):
        pass
    pdf.ln()
    lightblue = (173, 216, 230)
    with pdf.table(TABLE_DATA, cell_fill_color=lightblue, cell_fill_mode="COLUMNS"):
        pass
    assert_pdf_equal(pdf, HERE / "table_with_cell_fill.pdf", tmp_path)


class EvenOddCellFillMode:
    @staticmethod
    def should_fill_cell(i, j):
        return i % 2 and j % 2


def test_table_with_cell_fill_custom_class(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    lightblue = (173, 216, 230)
    with pdf.table(
        TABLE_DATA, cell_fill_color=lightblue, cell_fill_mode=EvenOddCellFillMode()
    ):
        pass
    assert_pdf_equal(pdf, HERE / "table_with_cell_fill_custom_class.pdf", tmp_path)


def test_table_with_internal_layout(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    with pdf.table(TABLE_DATA, borders_layout="INTERNAL"):
        pass
    assert_pdf_equal(pdf, HERE / "table_with_internal_layout.pdf", tmp_path)


def test_empty_table_with_internal_layout(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    with pdf.table(borders_layout="INTERNAL") as table:
        for y in range(3):
            row = table.row()
            for x in range(3):
                row.cell()
    assert_pdf_equal(pdf, HERE / "empty_table_with_internal_layout.pdf", tmp_path)


def test_table_with_minimal_layout(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    pdf.set_draw_color(100)  # dark grey
    pdf.set_line_width(1)
    with pdf.table(TABLE_DATA, borders_layout="MINIMAL"):
        pass
    assert_pdf_equal(pdf, HERE / "table_with_minimal_layout.pdf", tmp_path)


def test_table_with_single_top_line_layout(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    pdf.set_draw_color(50)  # very dark grey
    pdf.set_line_width(0.5)
    with pdf.table(TABLE_DATA, borders_layout="SINGLE_TOP_LINE"):
        pass
    assert_pdf_equal(pdf, HERE / "table_with_single_top_line_layout.pdf", tmp_path)


def test_table_with_single_top_line_layout_and_page_break(tmp_path):  # PR #912
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    pdf.set_draw_color(50)  # very dark grey
    pdf.set_line_width(0.5)
    data = list(MULTILINE_TABLE_DATA)
    # Reducing the text content on the 3rd line
    # so that there is an even number of rows on the 1st page:
    data[2] = (data[2][0][:-30], data[2][1])
    with pdf.table(
        borders_layout="SINGLE_TOP_LINE", cell_fill_color=230, cell_fill_mode="ROWS"
    ) as table:
        for data_row in data:
            row = table.row()
            for datum in data_row:
                row.cell(datum)
    assert_pdf_equal(
        pdf, HERE / "table_with_single_top_line_layout_and_page_break.pdf", tmp_path
    )


def test_table_with_page_break_and_headings_repeated(tmp_path):  # issue 1151
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    pdf.cell(text="repeat_headings=1:", new_y="NEXT")
    with pdf.table(
        MULTILINE_TABLE_DATA,
        repeat_headings=1,
    ):
        pass
    pdf.cell(text='repeat_headings="NONE":', new_y="NEXT")
    with pdf.table(
        MULTILINE_TABLE_DATA,
        repeat_headings="NONE",
    ):
        pass
    pdf.cell(text='repeat_headings="ON_TOP_OF_EVERY_PAGE":', new_y="NEXT")
    with pdf.table(
        MULTILINE_TABLE_DATA,
        repeat_headings="ON_TOP_OF_EVERY_PAGE",
    ):
        pass
    assert_pdf_equal(
        pdf, HERE / "table_with_page_break_and_headings_repeated.pdf", tmp_path
    )


def test_table_with_custom_border_layout(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font(family="helvetica", size=11)

    # generate a custom layout with most of the available features: changes in thickness, color,
    # and dash
    class CustomLayout(TableBordersLayout):
        def cell_style_getter(
            self,
            row_idx,
            col_idx,
            col_pos,
            num_heading_rows,
            num_rows,
            num_col_idx,
            num_col_pos,
        ):
            return TableCellStyle(
                left=(
                    TableBorderStyle(thickness=2)
                    if col_idx == 0
                    else TableBorderStyle(thickness=1.0, color=(0, 255, 125))
                ),
                bottom=(
                    TableBorderStyle(thickness=2) if row_idx == num_rows - 1 else False
                ),
                right=(
                    TableBorderStyle(thickness=2)
                    if col_idx == num_col_idx - 1
                    else False
                ),
                top=(
                    TableBorderStyle(thickness=2)
                    if row_idx in (0, num_heading_rows)
                    else (
                        True
                        if (row_idx - num_heading_rows) % 2 == 0
                        else TableBorderStyle(color=(255, 0, 0), dash=2)
                    )
                ),
            )

    with pdf.table(rows=TABLE_DATA, borders_layout=CustomLayout()):
        pass

    assert_pdf_equal(pdf, HERE / "table_with_custom_layout.pdf", tmp_path)


def test_table_align(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    with pdf.table(TABLE_DATA, text_align="CENTER"):
        pass
    pdf.ln()
    with pdf.table(TABLE_DATA, text_align=("CENTER", "CENTER", "RIGHT", "LEFT")):
        pass
    assert_pdf_equal(pdf, HERE / "table_align.pdf", tmp_path)


def test_table_capture_font_settings(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    black = (0, 0, 0)
    lightblue = (173, 216, 230)
    with pdf.table(headings_style=FontFace(color=black, emphasis="B")) as table:
        for row_num, data_row in enumerate(TABLE_DATA):
            with pdf.local_context(text_color=lightblue):
                row = table.row()
                for col_num, datum in enumerate(data_row):
                    font_style = FontFace(
                        emphasis="I" if row_num > 0 and col_num == 0 else None
                    )
                    row.cell(datum, style=font_style)
    assert_pdf_equal(pdf, HERE / "table_capture_font_settings.pdf", tmp_path)


def test_table_with_ttf_font(caplog, tmp_path):  # issue 749
    caplog.set_level(logging.ERROR)  # hides fonttool warnings
    pdf = FPDF()
    pdf.add_page()
    pdf.add_font(fname=HERE / "../fonts/cmss12.ttf")
    pdf.set_font("cmss12", size=16)
    with pdf.table(TABLE_DATA, first_row_as_headings=False):
        pass
    assert_pdf_equal(pdf, HERE / "table_with_ttf_font.pdf", tmp_path)


def test_table_with_ttf_font_and_headings(caplog, tmp_path):
    caplog.set_level(logging.ERROR)  # hides fonttool warnings
    pdf = FPDF()
    pdf.add_page()
    pdf.add_font("Roboto", fname=HERE / "../fonts/Roboto-Regular.ttf")
    pdf.add_font("Roboto", style="BI", fname=HERE / "../fonts/Roboto-BoldItalic.TTF")
    pdf.set_font("Roboto", size=16)
    with pdf.table(TABLE_DATA, headings_style=FontFace(emphasis="IB")):
        pass
    assert_pdf_equal(pdf, HERE / "table_with_ttf_font_and_headings.pdf", tmp_path)


def test_table_with_ttf_font_and_headings_but_missing_bold_font():
    pdf = FPDF()
    pdf.add_page()
    pdf.add_font("Quicksand", fname=HERE / "../fonts/Quicksand-Regular.otf")
    pdf.set_font("Quicksand", size=16)
    with pytest.raises(FPDFException) as error:
        with pdf.table(TABLE_DATA):
            pass
    assert (
        str(error.value)
        == "Using font 'quicksand' with emphasis 'B' in table headings require the corresponding font style to be added using add_font()"
    )


def test_table_with_cell_overflow(tmp_path):
    pdf = FPDF()
    pdf.set_font("Times", size=30)
    pdf.add_page()
    with pdf.table(width=pdf.epw / 2, col_widths=(1, 2, 1)) as table:
        row = table.row()
        row.cell("left")
        row.cell("center")
        row.cell("right")  # triggers header cell overflow on last column
        row = table.row()
        row.cell("A1")
        row.cell("A2")
        row.cell("A33333333")  # triggers cell overflow on last column
        row = table.row()
        row.cell("B1")
        row.cell("B2")
        row.cell("B3")
    assert_pdf_equal(pdf, HERE / "table_with_cell_overflow.pdf", tmp_path)


def test_table_with_gutter(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    with pdf.table(TABLE_DATA, gutter_height=3, gutter_width=3):
        pass
    pdf.ln(10)
    with pdf.table(
        TABLE_DATA, borders_layout="SINGLE_TOP_LINE", gutter_height=3, gutter_width=3
    ):
        pass
    assert_pdf_equal(pdf, HERE / "table_with_gutter.pdf", tmp_path)


def test_table_with_gutter_and_width(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    pdf.cell(text="Without gutter:", new_y="NEXT")
    pdf.ln()
    with pdf.table(TABLE_DATA, width=150):
        pass
    pdf.ln()
    pdf.cell(text="With gutter:", new_y="NEXT")
    pdf.ln()
    with pdf.table(TABLE_DATA, width=150, gutter_height=3, gutter_width=3):
        pass
    assert_pdf_equal(pdf, HERE / "table_with_gutter_and_width.pdf", tmp_path)


def test_table_with_colspan_and_gutter(tmp_path):  # issue 808
    pdf = FPDF()
    pdf.set_font("Times", size=30)
    pdf.add_page()
    with pdf.table(col_widths=(1, 2, 1, 1), gutter_height=5, gutter_width=5) as table:
        row = table.row()
        row.cell("0")
        row.cell("1")
        row.cell("2")
        row.cell("3")
        row = table.row()
        row.cell("A1")
        row.cell("A2", colspan=2)
        row.cell("A3")
        row = table.row()
        row.cell("B1", colspan=2)
        row.cell("B2")
        row.cell("B3")
    assert_pdf_equal(pdf, HERE / "table_with_colspan_and_gutter.pdf", tmp_path)


def test_table_with_capitalized_font_family_and_emphasis():  # issue 828
    pdf = FPDF()
    pdf.add_page()
    with pdf.table(
        TABLE_DATA, headings_style=FontFace(family="Helvetica", emphasis="ITALICS")
    ):
        pass


def test_table_with_no_headers_nor_horizontal_lines(tmp_path):  # discussion 924
    pdf = FPDF()
    pdf.set_font("Helvetica")
    pdf.add_page()
    with pdf.table(
        TABLE_DATA,
        cell_fill_color=200,
        cell_fill_mode="ROWS",
        text_align="LEFT",
        borders_layout="NO_HORIZONTAL_LINES",
        first_row_as_headings=False,
    ):
        pass
    assert_pdf_equal(
        pdf, HERE / "table_with_no_headers_nor_horizontal_lines.pdf", tmp_path
    )


def test_table_page_break_with_table_in_header(tmp_path):  # issue 943
    class PDF(FPDF):
        def header(self):
            with self.table() as t:
                r = t.row()
                r.cell("headertext")

    pdf = PDF()
    pdf.set_font("helvetica", style="B", size=8)
    pdf.add_page()
    with pdf.table() as table:
        for _ in range(1, 15):
            for data_row in TABLE_DATA:
                table.row(data_row)
    assert_pdf_equal(pdf, HERE / "table_page_break_with_table_in_header.pdf", tmp_path)


def test_table_with_multiple_headings_and_pagebreak(tmp_path):
    pdf = FPDF()
    pdf.set_font("Times", size=12)
    pdf.add_page()
    pdf.set_y(240)
    with pdf.table(
        num_heading_rows=2,
    ) as table:
        for j, rowdata in enumerate(MULTI_HEADING_TABLE_DATA):
            if j == 0:
                # row with colspan
                row = table.row()
                for cell in rowdata:
                    row.cell(text=cell, colspan=3)
            else:
                table.row(cells=rowdata)
    assert_pdf_equal(
        pdf,
        HERE / "table_with_multiple_headings_and_pagebreak.pdf",
        tmp_path,
    )


def test_table_num_heading_rows_and_first_row_as_headings():
    pdf = FPDF()
    pdf.set_font("Times", size=12)
    pdf.add_page()
    with pytest.raises(ValueError):
        with pdf.table(TABLE_DATA, first_row_as_headings=True, num_heading_rows=0):
            pass
    with pytest.raises(ValueError):
        with pdf.table(TABLE_DATA, first_row_as_headings=False, num_heading_rows=2):
            pass


def test_table_with_multiple_headings_and_no_horizontal_lines(tmp_path):
    pdf = FPDF()
    pdf.set_font("Times", size=12)
    pdf.add_page()
    with pdf.table(
        borders_layout="NO_HORIZONTAL_LINES",
        num_heading_rows=2,
    ) as table:
        for j, rowdata in enumerate(MULTI_HEADING_TABLE_DATA):
            if j == 0:
                # row with colspan
                row = table.row()
                for cell in rowdata:
                    row.cell(text=cell, colspan=3)
            else:
                table.row(cells=rowdata)
    assert_pdf_equal(
        pdf,
        HERE / "table_with_multiple_headings_and_no_horizontal_lines.pdf",
        tmp_path,
    )


def test_table_with_minimal_layout_and_multiple_headings(tmp_path):
    pdf = FPDF()
    pdf.set_font("Times", size=12)
    pdf.add_page()
    pdf.set_draw_color(100)  # dark grey
    pdf.set_line_width(1)
    with pdf.table(borders_layout="MINIMAL", num_heading_rows=2) as table:
        for j, rowdata in enumerate(MULTI_HEADING_TABLE_DATA):
            if j == 0:
                # row with colspan
                row = table.row()
                for cell in rowdata:
                    row.cell(text=cell, colspan=3)
            else:
                table.row(cells=rowdata)
    assert_pdf_equal(
        pdf, HERE / "table_with_minimal_layout_and_multiple_headings.pdf", tmp_path
    )


def test_table_with_single_top_line_layout_and_multiple_headings(tmp_path):
    pdf = FPDF()
    pdf.set_font("Times", size=12)
    pdf.add_page()
    with pdf.table(
        borders_layout="SINGLE_TOP_LINE",
        num_heading_rows=2,
    ) as table:
        for j, rowdata in enumerate(MULTI_HEADING_TABLE_DATA):
            if j == 0:
                # row with colspan
                row = table.row()
                for cell in rowdata:
                    row.cell(text=cell, colspan=3)
            else:
                table.row(cells=rowdata)

    assert_pdf_equal(
        pdf,
        HERE / "table_with_single_top_line_layout_and_multiple_headings.pdf",
        tmp_path,
    )


def test_table_with_no_horizontal_lines_layout(tmp_path):
    pdf = FPDF()
    pdf.set_font("Times", size=12)
    pdf.add_page()
    with pdf.table(
        TABLE_DATA,
        borders_layout="NO_HORIZONTAL_LINES",
        num_heading_rows=1,
    ):
        pass
    assert_pdf_equal(
        pdf,
        HERE / "table_with_no_horizontal_lines_layout.pdf",
        tmp_path,
    )


def test_table_with_heading_style_overrides(tmp_path):
    pdf = FPDF()
    pdf.set_font(family="helvetica", size=10)
    pdf.add_page()

    with pdf.table(
        headings_style=FontFace(emphasis="B", size_pt=18), num_heading_rows=2
    ) as table:
        # should be Helvetica bold size 18
        table.row().cell("Big Heading", colspan=3)
        second_header = table.row()
        # should be Helvetica bold size 14:
        second_header_style_1 = FontFace(size_pt=14)
        second_header.cell("First", style=second_header_style_1)
        # should be Times italic size 14
        second_header_style_2_3 = FontFace(family="times", emphasis="I", size_pt=14)
        second_header.cell("Second", style=second_header_style_2_3)
        second_header.cell("Third", style=second_header_style_2_3)
        # should be helvetica normal size 10
        table.row(("Some", "Normal", "Data"))

    assert_pdf_equal(pdf, HERE / "table_with_heading_style_overrides.pdf", tmp_path)


def test_table_with_set_fill_color(tmp_path):  # issue 963
    pdf = FPDF()
    pdf.set_font("helvetica", size=10)
    pdf.add_page()
    with pdf.table(first_row_as_headings=False) as table:
        row = table.row()
        pdf.set_fill_color(200)
        pdf.set_fill_color(200, 200, 200)
        row.cell("Hello")
    assert_pdf_equal(pdf, HERE / "table_with_set_fill_color.pdf", tmp_path)


def test_table_with_fill_color_set_beforehand(tmp_path):  # issue 932
    pdf = FPDF()
    pdf.set_font("Helvetica")
    pdf.set_fill_color((126, 217, 87))  # green
    pdf.add_page()
    with pdf.table(
        cell_fill_color=(200, 200, 200),  # light grey
        cell_fill_mode="COLUMNS",
        headings_style=FontFace(fill_color=(255, 255, 255)),  # white
    ) as table:
        for i, data_row in enumerate(TABLE_DATA):
            if i == 2:
                style = FontFace(fill_color=(250, 128, 114))  # salmon
            else:
                style = None
            row = table.row(style=style)
            for j, datum in enumerate(data_row):
                if i == 2 and j == 2:
                    style = FontFace(fill_color=(50, 50, 50))  # very dark grey
                else:
                    style = None
                row.cell(datum, style=style)
    assert_pdf_equal(pdf, HERE / "table_with_fill_color_set_beforehand.pdf", tmp_path)


def test_table_with_links(tmp_path):  # issue 1031
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=16)
    with pdf.table() as table:
        for i, data_row in enumerate(TABLE_DATA):
            row = table.row()
            for j, datum in enumerate(data_row):
                if j == 2 and i > 0:
                    row.cell(text=datum, link="https://py-pdf.github.io/fpdf2/")
                else:
                    row.cell(datum)
    assert_pdf_equal(pdf, HERE / "table_with_links.pdf", tmp_path)


def test_table_with_varying_col_count(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Helvetica")
    # test table with reducing number of rows
    with pdf.table() as table:
        for i, data_row in enumerate(TABLE_DATA):
            subset = data_row[:-i] if i else data_row
            table.row(subset)
    pdf.ln(3)
    # table with less columns in first row
    with pdf.table(first_row_as_headings=False) as table:
        for i, data_row in enumerate(TABLE_DATA[1:], start=1):
            subset = data_row[:i]
            table.row(subset)

    assert_pdf_equal(pdf, HERE / "table_with_varying_col_count.pdf", tmp_path)


def test_table_cell_fill_mode(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Helvetica")
    light_sky_blue = (150, 200, 255)
    headings_style = FontFace(fill_color=128)  # grey
    for mode in ("ROWS", "COLUMNS", "EVEN_ROWS", "EVEN_COLUMNS"):
        pdf.cell(text=f"cell_fill_mode=TableCellFillMode.{mode}:")
        pdf.ln(10)
        with pdf.table(
            TABLE_DATA,
            headings_style=headings_style,
            cell_fill_mode=getattr(TableCellFillMode, mode),
            cell_fill_color=light_sky_blue,
        ):
            pass
        pdf.ln()
    assert_pdf_equal(pdf, HERE / "table_cell_fill_mode.pdf", tmp_path)


def test_table_with_very_long_text():
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Helvetica")
    with pytest.raises(ValueError) as error:
        with pdf.table() as table:
            row = table.row()
            row.cell(LOREM_IPSUM)
            # Adding columns to have the content of the 1st cell to overflow:
            row.cell("")
            row.cell("")
    assert (
        str(error.value)
        == "The row with index 0 is too high and cannot be rendered on a single page"
    )
    with pytest.raises(ValueError) as error:
        with pdf.table() as table:
            row = table.row()
            row.cell("")
            row.cell("")
            row.cell(LOREM_IPSUM)
    assert (
        str(error.value)
        == "The row with index 0 is too high and cannot be rendered on a single page"
    )


def test_table_cell_border_none(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=20)
    with pdf.table(gutter_height=3, gutter_width=3) as table:
        for data_row in TABLE_DATA:
            row = table.row()
            for datum in data_row:
                row.cell(datum, border="None")
    assert_pdf_equal(pdf, HERE / "test_table_cell_border_none.pdf", tmp_path)


def test_table_cell_different_borders(tmp_path):
    pdf = FPDF()
    pdf.set_font("Times", size=20)
    pdf.add_page()
    pdf.cell(
        0,
        10,
        "border = left",
    )
    pdf.ln(20)
    with pdf.table(gutter_height=3, gutter_width=3) as table:
        for data_row in TABLE_DATA:
            row = table.row()
            for datum in data_row:
                row.cell(datum, border="left")
    pdf.add_page()
    pdf.cell(0, 10, "border = right")
    pdf.ln(20)
    with pdf.table(gutter_height=3, gutter_width=3) as table:
        for data_row in TABLE_DATA:
            row = table.row()
            for datum in data_row:
                row.cell(datum, border="right")
    pdf.add_page()
    pdf.cell(0, 10, "border = top")
    pdf.ln(20)
    with pdf.table(gutter_height=3, gutter_width=3) as table:
        for data_row in TABLE_DATA:
            row = table.row()
            for datum in data_row:
                row.cell(datum, border="top")
    pdf.add_page()
    pdf.cell(0, 10, "border = bottom")
    pdf.ln(20)
    with pdf.table(gutter_height=3, gutter_width=3) as table:
        for data_row in TABLE_DATA:
            row = table.row()
            for datum in data_row:
                row.cell(datum, border="bottom")
    pdf.add_page()
    pdf.cell(0, 10, "border = [left, right]")
    pdf.ln(20)
    with pdf.table(gutter_height=3, gutter_width=3) as table:
        for data_row in TABLE_DATA:
            row = table.row()
            for datum in data_row:
                row.cell(datum, border=3)
    pdf.add_page()
    pdf.cell(0, 10, "border =[left, top]")
    pdf.ln(20)
    with pdf.table(gutter_height=3, gutter_width=3) as table:
        for data_row in TABLE_DATA:
            row = table.row()
            for datum in data_row:
                row.cell(datum, border=5)
    pdf.add_page()
    pdf.cell(0, 10, "border =[left, bottom]")
    pdf.ln(20)
    with pdf.table(gutter_height=3, gutter_width=3) as table:
        for data_row in TABLE_DATA:
            row = table.row()
            for datum in data_row:
                row.cell(datum, border=9)
    pdf.add_page()
    pdf.cell(0, 10, "border =[right, top]")
    pdf.ln(20)
    with pdf.table(gutter_height=3, gutter_width=3) as table:
        for data_row in TABLE_DATA:
            row = table.row()
            for datum in data_row:
                row.cell(datum, border=6)
    pdf.add_page()
    pdf.cell(0, 10, "border =[right, bottom]")
    pdf.ln(20)
    with pdf.table(gutter_height=3, gutter_width=3) as table:
        for data_row in TABLE_DATA:
            row = table.row()
            for datum in data_row:
                row.cell(datum, border=10)
    pdf.add_page()
    pdf.cell(0, 10, "border =[top, bottom]")
    pdf.ln(20)
    with pdf.table(gutter_height=3, gutter_width=3) as table:
        for data_row in TABLE_DATA:
            row = table.row()
            for datum in data_row:
                row.cell(datum, border=12)
    pdf.add_page()
    pdf.cell(0, 10, "border =[left, right, top]")
    pdf.ln(20)
    with pdf.table(gutter_height=3, gutter_width=3) as table:
        for data_row in TABLE_DATA:
            row = table.row()
            for datum in data_row:
                row.cell(datum, border=7)
    pdf.add_page()
    pdf.cell(0, 10, "border =[left, right, bottom]")
    pdf.ln(20)
    with pdf.table(gutter_height=3, gutter_width=3) as table:
        for data_row in TABLE_DATA:
            row = table.row()
            for datum in data_row:
                row.cell(datum, border=11)
    pdf.add_page()
    pdf.cell(0, 10, "border =[left, right, top, bottom]")
    pdf.ln(20)
    with pdf.table(gutter_height=3, gutter_width=3) as table:
        for data_row in TABLE_DATA:
            row = table.row()
            for datum in data_row:
                row.cell(datum, border=15)
    assert_pdf_equal(pdf, HERE / "test_table_cell_different_borders.pdf", tmp_path)


def test_table_cell_border_all(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=20)
    with pdf.table(gutter_height=3, gutter_width=3) as table:
        for data_row in TABLE_DATA:
            row = table.row()
            for datum in data_row:
                print(datum, end=" ")
                row.cell(datum, border="all")
    assert_pdf_equal(pdf, HERE / "test_table_cell_border_all.pdf", tmp_path)


def test_table_cell_border_inherit(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=20)
    pdf.cell(0, 10, "border = 'inherit' and borders_layout = 'none'")
    pdf.ln(20)
    with pdf.table(borders_layout="none", gutter_height=3, gutter_width=3) as table:
        for data_row in TABLE_DATA:
            row = table.row()
            for datum in data_row:
                row.cell(datum, border="inherit")
    pdf.add_page()
    pdf.cell(0, 10, "border = 'inherit' and borders_layout = 'all'")
    pdf.ln(20)
    with pdf.table(borders_layout="all", gutter_height=3, gutter_width=3) as table:
        for data_row in TABLE_DATA:
            row = table.row()
            for datum in data_row:
                print(datum, end=" ")
                row.cell(datum, border="inherit")
    pdf.add_page()
    pdf.cell(0, 10, "border = 'inherit' and borders_layout = 'horizontal_lines'")
    pdf.ln(20)
    with pdf.table(
        borders_layout="horizontal_lines", gutter_height=3, gutter_width=3
    ) as table:
        for data_row in TABLE_DATA:
            row = table.row()
            for datum in data_row:
                row.cell(datum, border="inherit")
    pdf.add_page()
    pdf.cell(0, 10, "border = 'inherit' and borders_layout = 'internal' ")
    pdf.ln(20)
    with pdf.table(borders_layout="internal", gutter_height=3, gutter_width=3) as table:
        for data_row in TABLE_DATA:
            row = table.row()
            for datum in data_row:
                row.cell(datum, border="inherit")
    pdf.add_page()
    pdf.cell(0, 10, "border = 'inherit' and borders_layout = 'minimal'")
    pdf.ln(20)
    with pdf.table(borders_layout="minimal", gutter_height=3, gutter_width=3) as table:
        for data_row in TABLE_DATA:
            row = table.row()
            for datum in data_row:
                row.cell(datum, border="inherit")
    pdf.add_page()
    pdf.cell(0, 10, "border = 'inherit' and borders_layout = 'no_horizontal_lines'")
    pdf.ln(20)
    with pdf.table(
        borders_layout="no_horizontal_lines", gutter_height=3, gutter_width=3
    ) as table:
        for data_row in TABLE_DATA:
            row = table.row()
            for datum in data_row:
                row.cell(datum, border="inherit")
    pdf.add_page()
    pdf.cell(0, 10, "border = 'inherit' and borders_layout = 'single_top_line'")
    pdf.ln(20)
    with pdf.table(
        borders_layout="single_top_line", gutter_height=3, gutter_width=3
    ) as table:
        for data_row in TABLE_DATA:
            row = table.row()
            for datum in data_row:
                row.cell(datum, border="inherit")
    assert_pdf_equal(pdf, HERE / "test_table_cell_border_inherit.pdf", tmp_path)


def test_table_min_row_height(tmp_path):
    pdf = FPDF()
    pdf.add_page()
    pdf.set_font("Times", size=20)
    with pdf.table(min_row_height=30) as table:
        table.row(("A", "B"))
        table.row(("C", "D"), min_height=50)
    assert_pdf_equal(pdf, HERE / "table_min_row_height.pdf", tmp_path)
